#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import octobot_backtesting.api as backtesting_api

import octobot_commons.enums as common_enums

import octobot_services.api as service_api
import octobot_services.interfaces as service_interfaces
import octobot_services.managers as service_managers
import octobot_services.octobot_channel_consumer as service_channel_consumer

import octobot_tentacles_manager.api as tentacles_manager_api

import octobot.channels as octobot_channels
import octobot.constants as constants


class InterfaceProducer(octobot_channels.OctoBotChannelProducer):
    """Initializer class:
    - Initialize services, constants and tools
    """

    def __init__(self, channel, octobot):
        super().__init__(channel)
        self.octobot = octobot

        self.interfaces = []
        self.notifiers = []
        self.to_create_notifiers_count = 0

    async def start(self):
        in_backtesting = backtesting_api.is_backtesting_enabled(self.octobot.config)
        await self._create_interfaces(in_backtesting)
        await self._create_notifiers(in_backtesting)
        await self.start_interfaces()

    async def start_interfaces(self):
        to_start_interfaces = self.interfaces
        started_interfaces = await service_managers.start_interfaces(to_start_interfaces)
        if len(started_interfaces) != len(to_start_interfaces):
            missing_interfaces = [interface.get_name()
                                  for interface in to_start_interfaces
                                  if interface not in started_interfaces]
            self.logger.error(
                f"{', '.join(missing_interfaces)} interface{'s' if len(missing_interfaces) > 1 else ''} "
                f"did not start properly.")

    async def register_exchange(self, exchange_id):
        for to_notify_instance in self.interfaces + self.notifiers:
            await self._register_exchange(to_notify_instance, exchange_id)

    async def register_interface(self, instance):
        if instance is not None:
            self.interfaces.append(instance)
            await self._register_existing_exchanges(instance)

    async def register_notifier(self, instance):
        self.notifiers.append(instance)
        await self._register_existing_exchanges(instance)
        if len(self.notifiers) == self.to_create_notifiers_count:
            await service_api.process_pending_notifications()

    async def _register_existing_exchanges(self, instance):
        for exchange_id in self.octobot.exchange_producer.exchange_manager_ids:
            await self._register_exchange(instance, exchange_id)

    async def _create_interfaces(self, in_backtesting):
        # do not overwrite data in case of inner bots init (backtesting)
        if service_interfaces.get_bot_api() is None:
            service_api.initialize_global_project_data(self.octobot.octobot_api,
                                                       constants.PROJECT_NAME,
                                                       constants.LONG_VERSION)
        interface_factory = service_api.create_interface_factory(self.octobot.config)
        interface_list = interface_factory.get_available_interfaces()
        for interface_class in interface_list:
            await self._create_interface_if_relevant(interface_factory, interface_class, in_backtesting,
                                                     self.octobot.get_edited_config(constants.CONFIG_KEY,
                                                                                    dict_only=False))

    async def _create_notifiers(self, in_backtesting):
        notifier_factory = service_api.create_notifier_factory(self.octobot.config)
        notifier_list = notifier_factory.get_available_notifiers()
        for notifier_class in notifier_list:
            await self._create_notifier_class_if_relevant(notifier_factory, notifier_class, in_backtesting,
                                                          self.octobot.get_edited_config(constants.CONFIG_KEY,
                                                                                         dict_only=False))

    async def _create_interface_if_relevant(self, interface_factory, interface_class,
                                            backtesting_enabled, edited_config):
        if self._is_interface_relevant(interface_class, backtesting_enabled):
            await self.send(bot_id=self.octobot.bot_id,
                            subject=common_enums.OctoBotChannelSubjects.CREATION.value,
                            action=service_channel_consumer.OctoBotChannelServiceActions.INTERFACE.value,
                            data={
                                service_channel_consumer.OctoBotChannelServiceDataKeys.EDITED_CONFIG.value: 
                                    edited_config,
                                service_channel_consumer.OctoBotChannelServiceDataKeys.BACKTESTING_ENABLED.value:
                                    backtesting_enabled,
                                service_channel_consumer.OctoBotChannelServiceDataKeys.CLASS.value: interface_class,
                                service_channel_consumer.OctoBotChannelServiceDataKeys.FACTORY.value: interface_factory
                            })

    async def _create_notifier_class_if_relevant(self, notifier_factory, notifier_class,
                                                 backtesting_enabled, edited_config):
        if self._is_notifier_relevant(notifier_class, backtesting_enabled):
            await self.send(bot_id=self.octobot.bot_id,
                            subject=common_enums.OctoBotChannelSubjects.CREATION.value,
                            action=service_channel_consumer.OctoBotChannelServiceActions.NOTIFICATION.value,
                            data={
                                service_channel_consumer.OctoBotChannelServiceDataKeys.EDITED_CONFIG.value:
                                    edited_config,
                                service_channel_consumer.OctoBotChannelServiceDataKeys.BACKTESTING_ENABLED.value:
                                    backtesting_enabled,
                                service_channel_consumer.OctoBotChannelServiceDataKeys.CLASS.value: notifier_class,
                                service_channel_consumer.OctoBotChannelServiceDataKeys.FACTORY.value: notifier_factory,
                                service_channel_consumer.OctoBotChannelServiceDataKeys.EXECUTORS.value:
                                    self.octobot.task_manager.executors
                            })
            self.to_create_notifiers_count += 1

    async def _register_exchange(self, to_notify_instance, exchange_id):
        await self.send(bot_id=self.octobot.bot_id,
                        subject=common_enums.OctoBotChannelSubjects.UPDATE.value,
                        action=service_channel_consumer.OctoBotChannelServiceActions.EXCHANGE_REGISTRATION.value,
                        data={
                            service_channel_consumer.OctoBotChannelServiceDataKeys.INSTANCE.value: to_notify_instance,
                            service_channel_consumer.OctoBotChannelServiceDataKeys.EXCHANGE_ID.value: exchange_id,
                        })

    def _is_interface_relevant(self, interface_class, backtesting_enabled):
        return service_api.is_enabled(interface_class) and \
               tentacles_manager_api.is_tentacle_activated_in_tentacles_setup_config(
                   self.octobot.tentacles_setup_config,
                   interface_class.get_name()) and \
               all(service.get_is_enabled(self.octobot.config) for service in interface_class.REQUIRED_SERVICES) and \
               (not backtesting_enabled or (
                           backtesting_enabled and service_api.is_enabled_in_backtesting(interface_class)))

    def _is_notifier_relevant(self, notifier_class, backtesting_enabled):
        return service_api.is_enabled_in_config(notifier_class, self.octobot.config) and \
               tentacles_manager_api.is_tentacle_activated_in_tentacles_setup_config(
                   self.octobot.tentacles_setup_config,
                   notifier_class.get_name()) and \
               all(service.get_is_enabled(self.octobot.config)
                   for service in notifier_class.REQUIRED_SERVICES) and \
               not backtesting_enabled

    async def stop(self):
        self.logger.debug("Stopping ...")
        await service_managers.stop_interfaces(self.interfaces)
        self.logger.debug("Stopped")



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import time
import asyncio
import json
import requests
import threading

import octobot_commons.logging as logging
import octobot_commons.configuration as configuration
import octobot_commons.os_util as os_util
import octobot_commons.symbols.symbol_util as symbol_util
import octobot_commons.authentication as authentication

import octobot_commons.constants as common_constants

import octobot_evaluators.api as evaluator_api
import octobot_evaluators.enums as evaluator_enums

import octobot_services.constants as service_constants

import octobot_trading.api as trading_api

import octobot.community.models.community_fields as community_fields
import octobot.constants as constants


class CommunityManager:
    _headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}

    def __init__(self, octobot_api):
        self.octobot_api = octobot_api
        self.edited_config: configuration.Configuration = octobot_api.get_edited_config(dict_only=False)
        self.enabled = constants.IS_CLOUD_ENV or self.edited_config.get_metrics_enabled()
        self.reference_market = trading_api.get_reference_market(self.edited_config.config)
        self.logger = logging.get_logger(self.__class__.__name__)
        self.current_config = None
        self.keep_running = True
        self.session = octobot_api.get_aiohttp_session()
        try:
            self.bot_id = self.edited_config.get_metrics_id()
        except KeyError:
            self.bot_id = None

        # these attributes will be set at the last moment to ensure relevance and let time for everything to startup
        self.has_real_trader = None
        self.has_simulator = None
        self.exchange_managers = None

    def _init_community_config(self):
        self.has_real_trader = trading_api.is_trader_enabled_in_config(self.edited_config.config)
        self.has_simulator = trading_api.is_trader_simulator_enabled_in_config(self.edited_config.config)
        self.exchange_managers = trading_api.get_exchange_managers_from_exchange_ids(
            self.octobot_api.get_exchange_manager_ids())

    async def start_community_task(self):
        if self.enabled:
            try:
                # first ensure this session is not just a configuration test: register after a timer
                await asyncio.sleep(common_constants.TIMER_BEFORE_METRICS_REGISTRATION_SECONDS)
                self._init_community_config()
                await self.register_session()
                await self._update_authenticated_bot()
                while self.keep_running:
                    # send a keepalive at periodic intervals
                    await asyncio.sleep(common_constants.TIMER_BETWEEN_METRICS_UPTIME_UPDATE)
                    try:
                        await self._update_session()
                        await self._update_authenticated_bot()
                    except Exception as e:
                        self.logger.debug(f"Exception when handling community data : {e}")
            except asyncio.CancelledError:
                pass
            except Exception as e:
                self.logger.debug(f"Exception when handling community registration: {e}")

    async def stop_task(self):
        self.logger.debug("Stopping ...")
        self.keep_running = False
        await self.session.close()
        self.logger.debug("Stopped ...")

    @staticmethod
    def should_register_bot(config: configuration.Configuration):
        try:
            config.get_metrics_id()
            return True
        except KeyError:
            return False

    @staticmethod
    def background_get_id_and_register_bot(octobot_api):
        community_manager = CommunityManager(octobot_api)
        threading.Thread(target=community_manager._blocking_get_id_and_register, name="CommunityManagerGetId").start()

    def _blocking_get_id_and_register(self):
        try:
            resp = requests.get(f"{common_constants.METRICS_URL}{common_constants.METRICS_ROUTE_GEN_BOT_ID}",
                                headers=self._headers)
            text = resp.text
            if resp.status_code != 200:
                self.logger.debug(f"Impossible to get bot id: status code: {resp.status_code}, text: {text}")
            else:
                self.bot_id = json.loads(text)
                self._save_bot_id()
                community = self._get_bot_community()
                requests.post(f"{common_constants.METRICS_URL}{common_constants.METRICS_ROUTE_REGISTER}",
                              json=community, headers=self._headers)
        except Exception as e:
            self.logger.debug(f"Error when handling community: {e}")

    async def register_session(self, retry_on_error=True):
        self.current_config = await self._get_current_community_config()
        await self._post_community_data(common_constants.METRICS_ROUTE_REGISTER, self.current_config, retry_on_error)

    async def _update_session(self, retry_on_error=True):
        self.current_config[community_fields.CommunityFields.CURRENT_SESSION.value][
            community_fields.CommunityFields.UP_TIME.value] = int(time.time() - self.octobot_api.get_start_time())
        self.current_config[community_fields.CommunityFields.CURRENT_SESSION.value][
            community_fields.CommunityFields.PROFITABILITY.value] = self._get_profitability()
        self.current_config[community_fields.CommunityFields.CURRENT_SESSION.value][
            community_fields.CommunityFields.TRADED_VOLUMES.value] = self._get_traded_volumes()
        await self._post_community_data(common_constants.METRICS_ROUTE_UPTIME, self.current_config, retry_on_error)

    async def _update_authenticated_bot(self):
        try:
            if authentication.Authenticator.instance().is_logged_in():
                await authentication.Authenticator.instance().update_bot_config_and_stats(
                    self._get_profitability()
                )
        except Exception as err:
            self.logger.debug(f"Exception when pushing config and stats : {err}")

    async def _get_current_community_config(self):
        if not self.bot_id:
            await self._init_bot_id()
        if self.bot_id:
            return self._get_bot_community()

    def _get_bot_community(self):
        return {
            community_fields.CommunityFields.ID.value: self.bot_id,
            community_fields.CommunityFields.CURRENT_SESSION.value: {
                community_fields.CommunityFields.STARTED_AT.value: int(self.octobot_api.get_start_time()),
                community_fields.CommunityFields.UP_TIME.value: int(time.time() - self.octobot_api.get_start_time()),
                community_fields.CommunityFields.VERSION.value: constants.LONG_VERSION,
                community_fields.CommunityFields.SIMULATOR.value: self.has_simulator,
                community_fields.CommunityFields.TRADER.value: self.has_real_trader,
                community_fields.CommunityFields.EVAL_CONFIG.value: self._get_eval_config(),
                community_fields.CommunityFields.PAIRS.value: self._get_traded_pairs(),
                community_fields.CommunityFields.EXCHANGES.value: list(trading_api.get_exchange_names()),
                community_fields.CommunityFields.EXCHANGE_TYPES.value: self._get_exchange_types(),
                community_fields.CommunityFields.NOTIFICATIONS.value: self._get_notification_types(),
                community_fields.CommunityFields.TYPE.value: os_util.get_octobot_type(),
                community_fields.CommunityFields.PLATFORM.value: os_util.get_current_platform(),
                community_fields.CommunityFields.REFERENCE_MARKET.value: self.reference_market,
                community_fields.CommunityFields.PORTFOLIO_VALUE.value: self._get_real_portfolio_value(),
                community_fields.CommunityFields.PROFITABILITY.value: self._get_profitability(),
                community_fields.CommunityFields.TRADED_VOLUMES.value: self._get_traded_volumes(),
                community_fields.CommunityFields.SUPPORTS.value: self._get_supports(),
                community_fields.CommunityFields.SIGNAL_EMITTER.value:
                    authentication.Authenticator.instance().get_is_signal_emitter(),
                community_fields.CommunityFields.SIGNAL_RECEIVER.value:
                    authentication.Authenticator.instance().get_is_signal_receiver(),
                community_fields.CommunityFields.PROFILE_NAME.value: self.edited_config.profile.name,
                community_fields.CommunityFields.PROFILE_ID.value: self.edited_config.profile.profile_id,
                community_fields.CommunityFields.PROFILE_IMPORTED.value: self.edited_config.profile.imported,
            }
        }

    def _get_exchange_types(self):
        return [
            trading_api.get_exchange_type(exchange_manager).value
            for exchange_manager in self.exchange_managers
        ]

    def _get_profitability(self):
        total_origin_values = 0
        total_profitability = 0

        for exchange_manager in self.exchange_managers:
            profitability, _, _, _, _ = trading_api.get_profitability_stats(exchange_manager)
            total_profitability += float(profitability)
            total_origin_values += float(trading_api.get_origin_portfolio_value(exchange_manager))

        return total_profitability * 100 / total_origin_values if total_origin_values > 0 else 0

    def _get_traded_volumes(self):
        volume_by_currency = {}
        if self.has_real_trader:
            trades = []
            for exchange_manager in self.exchange_managers:
                trades += trading_api.get_trade_history(exchange_manager, since=self.octobot_api.get_start_time())
            for trade in trades:
                # cost is in quote currency for a traded pair
                currency = symbol_util.parse_symbol(trade.symbol).quote
                if currency in volume_by_currency:
                    volume_by_currency[currency] += float(trade.total_cost)
                else:
                    volume_by_currency[currency] = float(trade.total_cost)
        return volume_by_currency

    def _get_supports(self):
        supporting_exchanges = []
        for exchange_manager in self.exchange_managers:
            exchange_name = trading_api.get_exchange_name(exchange_manager)
            if self.has_real_trader \
               and trading_api.is_sponsoring(exchange_name) \
               and trading_api.is_valid_account(exchange_manager):
                supporting_exchanges.append(exchange_name)
        supports = authentication.Authenticator.instance().user_account.supports
        return {
            community_fields.CommunityFields.EXCHANGES.value: supporting_exchanges,
            community_fields.CommunityFields.ROLES.value: [supports.support_role],
            community_fields.CommunityFields.DONATIONS.value: [str(donation) for donation in supports.donations]
        }

    def _get_real_portfolio_value(self):
        if self.has_real_trader:
            total_value = 0
            for exchange_manager in self.exchange_managers:
                current_value = trading_api.get_current_portfolio_value(exchange_manager)
                # current_value might be 0 if no trades have been made / canceled => use origin value
                if current_value == 0:
                    current_value = trading_api.get_origin_portfolio_value(exchange_manager)
                total_value += current_value
            return float(total_value)
        else:
            return 0

    def _get_traded_pairs(self):
        pairs = set()
        for exchange_manager in self.exchange_managers:
            pairs = pairs.union(trading_api.get_trading_pairs(exchange_manager))
        return list(pairs)

    def _get_notification_types(self):
        has_notifications = service_constants.CONFIG_CATEGORY_NOTIFICATION in self.edited_config.config \
                            and service_constants.CONFIG_NOTIFICATION_TYPE in self.edited_config.config[
                                service_constants.CONFIG_CATEGORY_NOTIFICATION]
        return self.edited_config.config[service_constants.CONFIG_CATEGORY_NOTIFICATION][
            service_constants.CONFIG_NOTIFICATION_TYPE] if has_notifications else []

    def _get_eval_config(self):
        tentacle_setup_config = self.octobot_api.get_tentacles_setup_config()
        # trading mode
        config_eval = []
        if (trading_mode := self.octobot_api.get_trading_mode()) is not None:
            config_eval.append(trading_mode.get_name())

        # strategies
        for strategy in evaluator_api.get_evaluator_classes_from_type(
                evaluator_enums.EvaluatorMatrixTypes.STRATEGIES.value,
                tentacle_setup_config):
            config_eval.append(strategy.get_name())

        # evaluators
        evaluators = evaluator_api.get_evaluator_classes_from_type(evaluator_enums.EvaluatorMatrixTypes.TA.value,
                                                                   tentacle_setup_config)
        evaluators += evaluator_api.get_evaluator_classes_from_type(evaluator_enums.EvaluatorMatrixTypes.SOCIAL.value,
                                                                    tentacle_setup_config)
        evaluators += evaluator_api.get_evaluator_classes_from_type(
            evaluator_enums.EvaluatorMatrixTypes.REAL_TIME.value,
            tentacle_setup_config)
        for evaluator in evaluators:
            config_eval.append(evaluator.get_name())
        return config_eval

    async def _init_bot_id(self):
        try:
            async with self.session.get(f"{common_constants.METRICS_URL}{common_constants.METRICS_ROUTE_GEN_BOT_ID}",
                                        headers=self._headers) as resp:
                text = await resp.text()
                if resp.status != 200:
                    self.logger.debug(f"Impossible to get bot id: status code: {resp.status}, text: {text}")
                else:
                    self.bot_id = json.loads(text)
                    self._save_bot_id()
        except Exception as e:
            self.logger.debug(f"Error when handling community data : {e}")

    def _save_bot_id(self):
        if common_constants.CONFIG_METRICS not in self.edited_config.config \
                or not self.edited_config.config[common_constants.CONFIG_METRICS]:
            self.edited_config.config[common_constants.CONFIG_METRICS] = {common_constants.CONFIG_ENABLED_OPTION: True}
        self.edited_config.config[common_constants.CONFIG_METRICS][common_constants.CONFIG_METRICS_BOT_ID] = self.bot_id
        self.edited_config.save()

    async def _post_community_data(self, route, bot, retry_on_error):
        try:
            async with self.session.post(f"{common_constants.METRICS_URL}{route}", json=bot,
                                         headers=self._headers) as resp:
                await self._handle_post_error(resp, retry_on_error)
        except Exception as e:
            self.logger.debug(f"Error when handling community data : {e}")

    async def _handle_post_error(self, resp, retry_on_error):
        if resp.status != 200:
            if resp.status == 404:
                # did not found bot with id in config: generate new id and register new bot
                if retry_on_error:
                    await self._init_bot_id()
                    await self.register_session(retry_on_error=False)
            else:
                self.logger.debug(f"Impossible to send community data : "
                                  f"status code: {resp.status}, "
                                  f"text: {await resp.text()}")



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import time
import math

import octobot.constants as constants
import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.logging as logging
import octobot_commons.time_frame_manager as time_frame_manager
import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_evaluators.api as evaluators_api
import octobot_trading.api as trading_api


class ReachedLimitError(Exception):
    pass


def _apply_exchanges_limits(dict_config, logger, limit):
    if len(trading_api.get_enabled_exchanges_names(dict_config)) > limit:
        enabled_exchanges = []
        for exchange, config in dict_config.get(common_constants.CONFIG_EXCHANGES, {}).items():
            if config.get(common_constants.CONFIG_ENABLED_OPTION, True):
                if len(enabled_exchanges) < limit:
                    enabled_exchanges.append(exchange)
                else:
                    config[common_constants.CONFIG_ENABLED_OPTION] = False
                    logger.error("Disabled " + exchange)
        return f"Reached maximum allowed simultaneous exchanges for this plan, maximum is {limit}. " \
               f"Your OctoBot will trade on the following exchanges: {', '.join(enabled_exchanges)}"
    return ""


def _apply_symbols_limits(dict_config, logger, limit):
    enabled_symbols = []
    has_disabled_symbols = False
    message = ""
    for currency, crypto_currency_data in dict_config.get(common_constants.CONFIG_CRYPTO_CURRENCIES, {}).items():
        if crypto_currency_data.get(common_constants.CONFIG_ENABLED_OPTION, True):
            if len(enabled_symbols) >= limit:
                crypto_currency_data[common_constants.CONFIG_ENABLED_OPTION] = False
                logger.error(f"Disabled all {currency} trading pairs")
                has_disabled_symbols = True
                continue
            updated_symbols = []
            for symbol in crypto_currency_data.get(common_constants.CONFIG_CRYPTO_PAIRS, []):
                if symbol == common_constants.CONFIG_SYMBOLS_WILDCARD[0] \
                        or symbol == common_constants.CONFIG_SYMBOLS_WILDCARD:
                    crypto_currency_data[common_constants.CONFIG_ENABLED_OPTION] = False
                    message = f"Disabled wildcard symbol for {currency}. "
                    has_disabled_symbols = True
                    break
                else:
                    if len(enabled_symbols) < limit:
                        enabled_symbols.append(symbol)
                        updated_symbols.append(symbol)
                    else:
                        has_disabled_symbols = True
                        logger.error(f"Disabled {symbol} trading pair from {currency}")
            crypto_currency_data[common_constants.CONFIG_CRYPTO_PAIRS] = updated_symbols
    if has_disabled_symbols:
        return f"{message}Reached maximum allowed simultaneous trading pairs for this plan, maximum is {limit}. " \
               f"Your OctoBot will trade following pairs: {', '.join(enabled_symbols)}."
    return message


def _apply_time_frames_limits(full_config, logger, limit):
    tentacles_setup_config = tentacles_manager_api.get_tentacles_setup_config(full_config.get_tentacles_config_path())
    has_disabled_time_frames = False
    all_enabled_time_frames = []
    # patch time frames config
    for strategy_class in evaluators_api.get_activated_strategies_classes(tentacles_setup_config):
        config_time_frames = evaluators_api.get_time_frames_from_strategy(
            strategy_class, full_config.config, tentacles_setup_config
        )
        combined_time_frames = set(all_enabled_time_frames + config_time_frames)
        if len(combined_time_frames) < limit:
            all_enabled_time_frames = time_frame_manager.sort_time_frames(list(combined_time_frames))
        elif len(combined_time_frames) > limit:
            has_disabled_time_frames = True
            if len(all_enabled_time_frames) == limit:
                # no timeframe to add
                pass
            else:
                # disable shortest timeframes first
                missing_tf = time_frame_manager.sort_time_frames([
                    tf
                    for tf in config_time_frames
                    if tf not in all_enabled_time_frames
                ])
                added_time_frames = missing_tf[limit-len(all_enabled_time_frames):]
                all_enabled_time_frames = time_frame_manager.sort_time_frames(
                    list(all_enabled_time_frames) + added_time_frames
                )
        else:
            all_enabled_time_frames = list(combined_time_frames)
        if has_disabled_time_frames:
            should_update_config = False
            strategy_enabled_time_frames = [
                tf
                for tf in config_time_frames
                if tf in all_enabled_time_frames
            ]
            for time_frame in config_time_frames:
                if time_frame not in strategy_enabled_time_frames:
                    should_update_config = True
                    logger.error(f"Disabled {time_frame.value} time frame for {strategy_class.get_name()}")
            if should_update_config:
                evaluators_api.update_time_frames_config(
                    strategy_class, tentacles_setup_config, strategy_enabled_time_frames
                )
    if has_disabled_time_frames:
        return f"Reached maximum allowed simultaneous time frames for this plan, maximum is {limit}. " \
           f"Your OctoBot will trade using following time frames: " \
               f"{', '.join([tf.value for tf in all_enabled_time_frames])}."
    return ""


def apply_config_limits(configuration) -> list:
    logger = logging.get_logger("ConfigurationLimits")
    limit_warning_messages = []
    try:
        if constants.MAX_ALLOWED_EXCHANGES != constants.UNLIMITED_ALLOWED:
            if message := _apply_exchanges_limits(configuration.config, logger, constants.MAX_ALLOWED_EXCHANGES):
                limit_warning_messages.append(message)
        if constants.MAX_ALLOWED_SYMBOLS != constants.UNLIMITED_ALLOWED:
            if message := _apply_symbols_limits(configuration.config, logger, constants.MAX_ALLOWED_SYMBOLS):
                limit_warning_messages.append(message)
        if constants.MAX_ALLOWED_TIME_FRAMES != constants.UNLIMITED_ALLOWED:
            if message := _apply_time_frames_limits(configuration, logger, constants.MAX_ALLOWED_TIME_FRAMES):
                limit_warning_messages.append(message)
    except Exception as err:
        logger.exception(err, True, f"Error when applying limits: {err}")
    if limit_warning_messages:
        for message in limit_warning_messages:
            logger.error(message)
    return limit_warning_messages


def _check_max_backtesting_setting(setting_name, limit, values):
    if values and limit != constants.UNLIMITED_ALLOWED and len(values) > limit:
        raise ReachedLimitError(
            f"The maximum allowed simultaneous backtesting {setting_name} for your selected plan is {limit}"
        )


def _check_max_backtesting_candles_count(time_frames, start_timestamp, end_timestamp):
    time_frames = time_frames or [tf for tf in common_enums.TimeFrames]
    _check_max_backtesting_setting("time frames using custom duration", constants.MAX_ALLOWED_TIME_FRAMES, time_frames)
    if start_timestamp is None or constants.MAX_ALLOWED_BACKTESTING_CANDLES_HISTORY == constants.UNLIMITED_ALLOWED:
        return
    if end_timestamp is None:
        end_timestamp = time.time()
    shortest_time_frame = time_frame_manager.sort_time_frames(time_frames)[0]
    time_frame_seconds = common_enums.TimeFramesMinutes[shortest_time_frame] * common_constants.MINUTE_TO_SECONDS
    candles_count = math.floor((end_timestamp - start_timestamp) / time_frame_seconds)
    if candles_count > constants.MAX_ALLOWED_BACKTESTING_CANDLES_HISTORY:
        raise ReachedLimitError(
            f"For this plan, the maximum allowed backtesting candles per time frame "
            f"is {constants.MAX_ALLOWED_BACKTESTING_CANDLES_HISTORY}. "
            f"With the selected backtesting duration, the {shortest_time_frame.value} time frame would "
            f"cover {candles_count} candles. Please select other time frames or reduce the backtesting duration."
        )


def ensure_backtesting_limits(exchanges, symbols, time_frames, start_timestamp, end_timestamp) -> None:
    _check_max_backtesting_setting("exchanges", constants.MAX_ALLOWED_EXCHANGES, exchanges)
    _check_max_backtesting_setting("trading pairs", constants.MAX_ALLOWED_SYMBOLS, symbols)
    _check_max_backtesting_candles_count(time_frames, start_timestamp, end_timestamp)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import enum
import json
import os
import stat

import aiofiles
import aiohttp

import octobot.commands as commands
import octobot.constants as constants
import octobot.updater.updater as updater_class
import octobot_commons.aiohttp_util as aiohttp_util
import octobot_commons.constants as commons_constants
import octobot_commons.enums as commons_enums
import octobot_commons.os_util as os_util


class DeliveryPlatformsName(enum.Enum):
    WINDOWS = "windows"
    LINUX = "linux"
    MAC = "osx"


class DeliveryPlatformsExtension(enum.Enum):
    WINDOWS = ".exe"
    LINUX = ""
    MAC = ""


class DeliveryArchitectureName(enum.Enum):
    X64_X86 = "x64"
    ARM_X64 = "arm_x64"


class BinaryUpdater(updater_class.Updater):
    BINARY_DELIVERY_SEPARATOR = "_"
    OLD_BINARY_SUFFIX = ".bak"
    NEW_BINARY_SUFFIX = ".new"

    async def get_latest_version(self):
        return self._parse_latest_version(await self._get_latest_release_data())

    async def update_impl(self) -> bool:
        # TODO fix binary updater to use release endpoint (assents can't be downloaded from gh directly)
        self.logger.error(f"Please manually update your OctoBot executable from this page: "
                          f"{self._get_latest_release_url(False)}")
        return False
        new_binary_file = await self._download_binary()
        if new_binary_file is not None:
            self._give_execution_rights(new_binary_file)
            self._move_binaries(commands.get_bot_file(), new_binary_file)
            return True
        return False

    def _get_latest_release_url(self, use_api_url):
        base = f"{commons_constants.GITHUB_API_CONTENT_URL}/repos" if use_api_url else commons_constants.GITHUB_BASE_URL
        return f"{base}/{commons_constants.GITHUB_ORGANISATION}/{constants.PROJECT_NAME}/releases/latest"

    async def _get_latest_release_data(self):
        try:
            async with aiohttp.ClientSession().get(self._get_latest_release_url(True)) as resp:
                text = await resp.text()
                if resp.status == 200:
                    return json.loads(text)
            return None
        except Exception as e:
            self.logger.debug(f"Error when fetching latest binary data : {e}")

    def _parse_latest_version(self, latest_release_data):
        try:
            if latest_release_data.get("draft", False) or latest_release_data.get("prerelease", False):
                return None
            else:
                return latest_release_data["tag_name"]
        except (KeyError, AttributeError) as e:
            self.logger.debug(f"Error when parsing latest binary version : {e}")
        return None

    async def _get_asset_from_release_data(self):
        latest_release_data = await self._get_latest_release_data()
        release_asset_name = self._create_release_asset_name(os_util.get_os())
        return release_asset_name, self._get_asset_from_name(latest_release_data, release_asset_name)

    def _get_asset_from_name(self, release_data, expected_asset_name):
        try:
            for asset in release_data["assets"]:
                if expected_asset_name == asset["name"]:
                    return asset
        except KeyError as e:
            self.logger.debug(f"Error when searching for {expected_asset_name} in latest release data : {e}")
        return None

    def _create_release_asset_name(self, platform):
        if platform is commons_enums.PlatformsName.WINDOWS:
            if os_util.is_machine_64bit():
                return f"{constants.PROJECT_NAME}{self.BINARY_DELIVERY_SEPARATOR}" \
                       f"{DeliveryPlatformsName.WINDOWS.value}{self.BINARY_DELIVERY_SEPARATOR}" \
                       f"{DeliveryArchitectureName.X64_X86.value}{DeliveryPlatformsExtension.WINDOWS.value}"
        elif platform is commons_enums.PlatformsName.MAC:
            if os_util.is_machine_64bit():
                if os_util.is_machine_64bit():
                    return f"{constants.PROJECT_NAME}{self.BINARY_DELIVERY_SEPARATOR}" \
                           f"{DeliveryPlatformsName.MAC.value}{self.BINARY_DELIVERY_SEPARATOR}" \
                           f"{DeliveryArchitectureName.X64_X86.value}{DeliveryPlatformsExtension.MAC.value}"
        elif platform is commons_enums.PlatformsName.LINUX:
            if os_util.is_machine_64bit():
                return f"{constants.PROJECT_NAME}{self.BINARY_DELIVERY_SEPARATOR}" \
                       f"{DeliveryPlatformsName.LINUX.value}{self.BINARY_DELIVERY_SEPARATOR}" \
                       f"{DeliveryArchitectureName.X64_X86.value}{DeliveryPlatformsExtension.LINUX.value}"
            elif os_util.is_arm_machine():
                return f"{constants.PROJECT_NAME}{self.BINARY_DELIVERY_SEPARATOR}" \
                       f"{DeliveryPlatformsName.LINUX.value}{self.BINARY_DELIVERY_SEPARATOR}" \
                       f"{DeliveryArchitectureName.ARM_X64.value}{DeliveryPlatformsExtension.LINUX.value}"
        return None

    async def _download_binary(self):
        release_asset_name, matching_asset_binary = await self._get_asset_from_release_data()
        new_binary_file = f"{release_asset_name}{self.NEW_BINARY_SUFFIX}"
        new_binary_file_url = matching_asset_binary["browser_download_url"]
        if matching_asset_binary is None:
            self.logger.error(f"Error when downloading latest version binary : Release not found on server")
            return None
        self.logger.info(f"Start downloading OctoBot update at {new_binary_file_url}")
        async with aiofiles.open(new_binary_file, 'wb+') as downloaded_file:
            await aiohttp_util.download_stream_file(output_file=downloaded_file,
                                                    file_url=new_binary_file_url,
                                                    aiohttp_session=aiohttp.ClientSession(),
                                                    is_aiofiles_output_file=True)
        self.logger.info(f"OctoBot update downloaded successfully")
        return new_binary_file

    def _move_binaries(self, current_binary_file, new_binary_file):
        self.logger.info(f"Updating local binary file")
        old_binary_path = f"{current_binary_file}{self.OLD_BINARY_SUFFIX}"
        try:
            os.remove(old_binary_path)
        except OSError:
            self.logger.debug(f"{old_binary_path} doesn't exist")
        os.rename(current_binary_file, old_binary_path)
        os.rename(new_binary_file, current_binary_file)

    def _give_execution_rights(self, new_binary_file):
        if os_util.get_os() is not commons_enums.PlatformsName.WINDOWS:
            self.logger.info(f"Adding execution rights to updated OctoBot binary")
            st = os.stat(new_binary_file)
            os.chmod(new_binary_file, st.st_mode | stat.S_IEXEC)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import random
import time

import websockets
import asyncio
import enum
import json
import packaging.version as packaging_version

import octobot_commons.errors as commons_errors
import octobot_commons.enums as commons_enums
import octobot_commons.authentication as authentication
import octobot.constants as constants
import octobot.community.feeds.abstract_feed as abstract_feed
import octobot.community.identifiers_provider as identifiers_provider


class COMMANDS(enum.Enum):
    SUBSCRIBE = "subscribe"
    MESSAGE = "message"


class CHANNELS(enum.Enum):
    MESSAGE = "Spree::MessageChannel"


class CommunityWSFeed(abstract_feed.AbstractFeed):
    INIT_TIMEOUT = 60
    RECONNECT_DELAY = 15
    STALE_CONNECTION_TIMEOUT = 30   # ws server sends a ping every 3s

    def __init__(self, feed_url, authenticator):
        super().__init__(feed_url, authenticator)
        self.websocket_connection = None
        self.lock = asyncio.Lock()

        self.consumer_task = None
        self.watcher_task = None
        self._identifier_by_stream_id = {}
        self._reconnect_attempts = 0
        self._last_ping_time = None

    async def start(self):
        await self._ensure_connection()
        if self.consumer_task is None or self.consumer_task.done():
            self.consumer_task = asyncio.create_task(self.start_consumer())
        if self.watcher_task is None or self.watcher_task.done():
            self.watcher_task = asyncio.create_task(self.connection_watcher())

    async def stop(self):
        self.logger.debug("Stopping ...")
        self.should_stop = True
        if self.websocket_connection is not None:
            await self.websocket_connection.close()
        if self.consumer_task is not None:
            self.consumer_task.cancel()
        if self.watcher_task is not None:
            self.watcher_task.cancel()
        self.logger.debug("Stopped")

    # pylint: disable=E1101
    async def start_consumer(self):
        while not self.should_stop:
            try:
                await self._ensure_connection()
                async for message in self.websocket_connection:
                    try:
                        await self.consume(message)
                    except Exception as e:
                        self.logger.exception(e, True, f"Error while consuming feed: {e}")
            except (websockets.ConnectionClosedError, ConnectionRefusedError):
                if self._reconnect_attempts == 0:
                    self.logger.warning(f"Disconnected from community, reconnecting now.")
                    # first reconnect instantly
                else:
                    # use self.RECONNECT_DELAY + or - 0-10 %
                    reconnect_delay = self.RECONNECT_DELAY * (1 + (random.random() * 2 - 1) / 10)
                    self.logger.warning(f"Disconnected from community, retrying to connect in {reconnect_delay}s "
                                        f"({self._reconnect_attempts} attempts)")
                    if reconnect_delay > 0:
                        await asyncio.sleep(reconnect_delay)
            except Exception as e:
                self.logger.exception(e, True, f"Unexpected exception when receiving community feed: {e}")

    async def connection_watcher(self):
        while not self.should_stop:
            await asyncio.sleep(self.STALE_CONNECTION_TIMEOUT)
            if self._last_ping_time is not None and \
                    time.time() - self._last_ping_time > self.STALE_CONNECTION_TIMEOUT:
                self.logger.error("Stale community connection detected, resetting the connection")
                # close websocket_connection, the next _ensure_connection call will restart it
                await self.websocket_connection.close()

    async def consume(self, message):
        if message.startswith('{"type":"ping"'):
            self._last_ping_time = time.time()
            return
        parsed_message = json.loads(message)["message"]
        try:
            self._ensure_supported(parsed_message)
            for callback in self._get_callbacks(parsed_message):
                await callback(parsed_message)
        except commons_errors.UnsupportedError as e:
            self.logger.error(f"Unsupported message: {e}")

    def _ensure_supported(self, parsed_message):
        if packaging_version.Version(parsed_message[commons_enums.CommunityFeedAttrs.VERSION.value]) \
                < packaging_version.Version(constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION):
            raise commons_errors.UnsupportedError(
                f"Minimum version: {constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION}"
            )

    async def send(self, message, channel_type, identifier,
                   command=COMMANDS.MESSAGE.value, reconnect_if_necessary=True):
        if reconnect_if_necessary:
            await self._ensure_connection()
        if identifier is not None:
            await self._ensure_stream_identifier(identifier)
        await self.websocket_connection.send(self._build_ws_message(message, channel_type, command, identifier))

    def _build_ws_message(self, message, channel_type, command, identifier):
        return json.dumps({
            "command": command,
            "identifier": self._build_channel_identifier(),
            "data": self._build_data(channel_type, identifier, message)
        })

    def _build_data(self, channel_type, identifier, message):
        if message:
            return json.dumps({
                "topic": channel_type.value,
                "feed_id": self._build_stream_id(identifier),
                "version": constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION,
                "value": json.dumps(message),
            })
        return {}

    async def register_feed_callback(self, channel_type, callback, identifier=None):
        """
        Registers a feed callback
        """
        await self._ensure_stream_identifier(identifier)
        if identifier not in list(self._identifier_by_stream_id.values()):
            stream_id = await self._fetch_stream_identifier(identifier)
            self._identifier_by_stream_id[stream_id] = identifier
        try:
            self.feed_callbacks[channel_type][identifier].append(callback)
        except KeyError:
            if channel_type not in self.feed_callbacks:
                self.feed_callbacks[channel_type] = {}
            self.feed_callbacks[channel_type][identifier] = [callback]

    async def _ensure_stream_identifier(self, identifier):
        if identifier not in list(self._identifier_by_stream_id.values()):
            stream_id = await self._fetch_stream_identifier(identifier)
            self._identifier_by_stream_id[stream_id] = identifier

    async def _fetch_stream_identifier(self, identifier):
        if identifier is None:
            return None
        async with self.authenticator.get_aiohttp_session().get(
                f"{identifiers_provider.IdentifiersProvider.COMMUNITY_URL}api/v2/storefront/feeds/id/{identifier}"
        ) as resp:
            return (await resp.json())["feed_id"]

    def _get_callbacks(self, parsed_message):
        channel_type = self._get_channel_type(parsed_message)
        for callback in self.feed_callbacks.get(channel_type, {}).get(None, ()):
            yield callback
        try:
            identifier = self._get_identifier(parsed_message)
        except KeyError:
            self.logger.debug(f"Unknown feed identifier: "
                              f"{parsed_message[commons_enums.CommunityFeedAttrs.STREAM_ID.value]}")
            return
        if identifier is None:
            # do not yield the same callback twice
            return
        for callback in self.feed_callbacks.get(channel_type, {}).get(identifier, ()):
            yield callback

    def _get_channel_type(self, message):
        return commons_enums.CommunityChannelTypes(message[commons_enums.CommunityFeedAttrs.CHANNEL_TYPE.value])

    def _get_identifier(self, message):
        return self._identifier_by_stream_id[message[commons_enums.CommunityFeedAttrs.STREAM_ID.value]]

    def _build_channel_identifier(self):
        return json.dumps({
            "channel": CHANNELS.MESSAGE.value
        })

    def _build_stream_id(self, requested_identifier):
        for stream_id, identifier in self._identifier_by_stream_id.items():
            if requested_identifier == identifier:
                return stream_id
        return None

    async def _subscribe(self):
        await self.send({}, None, None, command=COMMANDS.SUBSCRIBE.value)
        # waiting for subscription confirmation
        try:
            await asyncio.wait_for(self._get_subscribe_answer(), self.INIT_TIMEOUT)
        except asyncio.TimeoutError:
            raise authentication.AuthenticationError(f"Failed to subscribe to feed")

    async def _get_subscribe_answer(self):
        async for message in self.websocket_connection:
            self.logger.debug("Waiting for subscription confirmation...")
            resp = json.loads(message)
            # TODO handle subscribe errors
            if resp.get("type") and resp.get("type") == "confirm_subscription":
                self.subscribed = True
                return

    # pylint: disable=E1101
    async def _ensure_connection(self):
        if not self.is_connected():
            async with self.lock:
                if not self.is_connected():
                    # (re)connect websocket
                    self._reconnect_attempts += 1
                    await self._connect()
                    self._reconnect_attempts = 0

    async def _connect(self):
        self.subscribed = False
        if self.authenticator.initialized_event is not None:
            await asyncio.wait_for(self.authenticator.initialized_event.wait(), self.INIT_TIMEOUT)
        if not self.authenticator.is_logged_in():
            raise authentication.AuthenticationRequired("OctoBot Community authentication is required to "
                                                        "use community trading signals")
        self.websocket_connection = await websockets.connect(self.feed_url,
                                                             extra_headers=self.authenticator.get_backend_headers(),
                                                             ping_interval=None)
        await self._subscribe()
        self.logger.info("Connected to community feed")

    def is_connected(self):
        return self.websocket_connection is not None and self.websocket_connection.open



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import time

import octobot_commons.logging as bot_logging


class AbstractFeed:
    def __init__(self, feed_url, authenticator):
        self.logger: bot_logging.BotLogger = bot_logging.get_logger(
            self.__class__.__name__
        )
        self.feed_url = feed_url
        self.should_stop = False
        self.authenticator = authenticator
        self.feed_callbacks = {}
        self.subscribed = False
        self.last_message_time = None
        self.is_signal_receiver = False
        self.is_signal_emitter = False

    async def start(self):
        raise NotImplementedError("start is not implemented")

    async def stop(self):
        raise NotImplementedError("stop is not implemented")

    async def register_feed_callback(self, channel_type, callback, identifier=None):
        raise NotImplementedError("register_feed_callback is not implemented")

    async def send(self, message, channel_type, identifier, **kwargs):
        raise NotImplementedError("send is not implemented")

    def can_connect(self):
        return True

    def is_connected_to_remote_feed(self):
        return False

    def update_last_message_time(self):
        self.last_message_time = time.time()

    def is_up_to_date_with_account(self, user_account):
        return True

    def is_connected(self):
        return False

    def remove_device_details(self):
        pass



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import contextlib
import copy
import typing
import storage3
import storage3.constants
import gotrue
import gotrue.errors
import postgrest
import supabase.lib.client_options

import octobot.community.supabase_backend.postgres_functions as postgres_functions
import octobot.community.supabase_backend.supabase_realtime_client as supabase_realtime_client


class AuthenticatedAsyncSupabaseClient(supabase.Client):
    """
    supabase.Client subclass to handle:
    - authenticated calls
    - database functions calls
    - auth token refresh
    - realtime client
    There should not be OctoBot specific code here
    """
    REQUEST_TIMEOUT = 20    # default timeout is 5 which is sometimes not enough
    def __init__(
        self,
        supabase_url: str,
        supabase_key: str,
        options: supabase.lib.client_options.ClientOptions = supabase.lib.client_options.ClientOptions(),
        loop=None,
    ):
        self.options = options
        self.options.postgrest_client_timeout = self.REQUEST_TIMEOUT
        self.auth: gotrue.SyncGoTrueClient = None
        self.postgrest: postgrest.AsyncPostgrestClient = None
        super().__init__(supabase_url, supabase_key, options=options)
        self.realtime: supabase_realtime_client.AuthenticatedSupabaseRealtimeClient = self._init_realtime_client(
            self.realtime_url,
            self.supabase_key,
        )
        # update postgres authentication upon auth state change
        self.auth.on_auth_state_change(self._use_auth_session)
        self.event_loop = None

    @staticmethod
    def _init_postgrest_client(
        rest_url: str,
        supabase_key: str,
        headers: typing.Dict[str, str],
        timeout,    # skip typing to avoid httpx import
        schema: str = "public",
    ) -> postgrest.AsyncPostgrestClient:
        """Private helper for creating an instance of the Postgrest client."""
        # Override to use postgrest.AsyncPostgrestClient and allow async requests
        client = postgrest.AsyncPostgrestClient(
            rest_url, headers=headers, schema=schema, timeout=timeout
        )
        client.auth(token=supabase_key)
        return client

    @staticmethod
    def _init_realtime_client(
        realtime_url: str, supabase_key: str
    ) -> supabase_realtime_client.AuthenticatedSupabaseRealtimeClient:
        return supabase_realtime_client.AuthenticatedSupabaseRealtimeClient(realtime_url, supabase_key)

    @staticmethod
    def _init_storage_client(
        storage_url: str,
        headers: typing.Dict[str, str],
        storage_client_timeout: int = storage3.constants.DEFAULT_TIMEOUT,
    ) -> storage3.AsyncStorageClient:
        return storage3.AsyncStorageClient(storage_url, headers, storage_client_timeout)

    def table(self, table_name: str) -> postgrest.AsyncRequestBuilder:  # typing override
        """Perform a table operation.

        Note that the supabase client uses the `from` method, but in Python,
        this is a reserved keyword, so we have elected to use the name `table`.
        Alternatively you can use the `.from_()` method.
        """
        return self.from_(table_name)

    @contextlib.asynccontextmanager
    async def other_postgres_client(self, supabase_url: str = None, supabase_key: str = None, schema: str = "public"):
        other_postgres = None
        try:
            other_postgres = await self.init_other_postgrest_client(
                supabase_url=supabase_url, supabase_key=supabase_key, schema=schema
            )
            yield other_postgres
        finally:
            if other_postgres is not None:
                await other_postgres.aclose()

    async def init_other_postgrest_client(
        self, supabase_url: str = None, supabase_key: str = None, schema: str = "public"
    ) -> postgrest.AsyncPostgrestClient:
        supabase_key = supabase_key or self.supabase_key
        headers = self.options.headers
        if supabase_key != self.supabase_key:
            headers = copy.deepcopy(postgrest.constants.DEFAULT_POSTGREST_CLIENT_HEADERS)
            headers.update(self._format_auth_headers(supabase_key, supabase_key))
        return AuthenticatedAsyncSupabaseClient._init_postgrest_client(
            rest_url=f"{supabase_url}/rest/v1" if supabase_url else self.rest_url,
            supabase_key=supabase_key,
            headers=headers,
            timeout=self.options.postgrest_client_timeout,
            schema=schema,
        )

    async def close(self):
        # timer has to be stopped, there is no public stop api
        if self.auth._refresh_token_timer:
            self.auth._refresh_token_timer.cancel()
            self.auth._refresh_token_timer = None
        try:
            await self.postgrest.aclose()
        except RuntimeError:
            # happens when the event loop is closed already
            pass
        await self.realtime.close()

    def postgres_functions(self):
        return postgres_functions.PostgresFunctions(self.supabase_url, self._get_auth_headers())

    def remove_session_details(self):
        self.auth._remove_session()
        self.auth._notify_all_subscribers("SIGNED_OUT", None)

    def _use_auth_session(self, event: gotrue.AuthChangeEvent, _):
        if event in (
            "SIGNED_IN",
            "SIGNED_OUT",
            "TOKEN_REFRESHED"
        ):
            session = self._get_auth_session()
            auth_token = session.access_token if session else None
            if auth_token:
                self.postgrest.auth(auth_token)
            self.realtime.set_auth(auth_token, self.event_loop)
            # todo uncomment when storage client has to be used by authenticated users instead of anon
            # await self.storage.aclose()
            # self._init_storage_client(self.storage_url, self._get_auth_headers())

    def _get_auth_headers(self):
        """Helper method to get auth headers."""
        # What's the corresponding method to get the token
        return self._format_auth_headers(self.supabase_key, self._get_auth_key())

    def _format_auth_headers(self, supabase_key, auth_token):
        return {
            "apiKey": supabase_key,
            "Authorization": f"Bearer {auth_token}",
        }

    def _get_auth_key(self):
        return self.supabase_key if self._get_auth_session() is None else self._get_auth_session().access_token

    def _get_auth_session(self):
        if self.auth is None:
            # return None when no use is signed in
            return None
        try:
            return self.auth.get_session()
        except gotrue.errors.AuthError:
            return None



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import octobot.constants as constants
import octobot.enums as enums
import octobot_commons.logging as logging
import octobot_commons.configuration as configuration


class IdentifiersProvider:
    ENABLED_ENVIRONMENT: str = None
    COMMUNITY_LANDING_URL: str = None
    COMMUNITY_URL: str = None
    FRONTEND_PASSWORD_RECOVER_URL: str = None
    BACKEND_URL: str = None
    BACKEND_KEY: str = None

    @staticmethod
    def use_production():
        IdentifiersProvider.COMMUNITY_URL = constants.OCTOBOT_COMMUNITY_URL
        IdentifiersProvider.COMMUNITY_LANDING_URL = constants.OCTOBOT_COMMUNITY_LANDING_URL
        IdentifiersProvider.FRONTEND_PASSWORD_RECOVER_URL = constants.OCTOBOT_COMMUNITY_RECOVER_PASSWORD_URL
        IdentifiersProvider.BACKEND_URL = constants.COMMUNITY_BACKEND_URL
        IdentifiersProvider.BACKEND_KEY = constants.COMMUNITY_BACKEND_KEY
        IdentifiersProvider._register_environment(enums.CommunityEnvironments.Production)

    @staticmethod
    def use_staging():
        IdentifiersProvider.COMMUNITY_URL = constants.STAGING_OCTOBOT_COMMUNITY_URL
        IdentifiersProvider.COMMUNITY_LANDING_URL = constants.STAGING_OCTOBOT_COMMUNITY_LANDING_URL
        IdentifiersProvider.FRONTEND_PASSWORD_RECOVER_URL = constants.STAGING_COMMUNITY_RECOVER_PASSWORD_URL
        IdentifiersProvider.BACKEND_URL = constants.STAGING_COMMUNITY_BACKEND_URL
        IdentifiersProvider.BACKEND_KEY = constants.STAGING_COMMUNITY_BACKEND_KEY
        IdentifiersProvider._register_environment(enums.CommunityEnvironments.Staging)

    @staticmethod
    def _register_environment(env):
        if IdentifiersProvider.ENABLED_ENVIRONMENT != env:
            logging.get_logger(IdentifiersProvider.__name__).debug(f"Using {env.value} Community environment.")
        IdentifiersProvider.ENABLED_ENVIRONMENT = env

    @staticmethod
    def use_default():
        if constants.USE_BETA_EARLY_ACCESS:
            IdentifiersProvider.use_staging()
        else:
            IdentifiersProvider.use_production()

    @staticmethod
    def is_staging_environment_enabled(config: dict):
        try:
            env = config[constants.CONFIG_COMMUNITY][constants.CONFIG_COMMUNITY_ENVIRONMENT]
            return enums.CommunityEnvironments(env) is enums.CommunityEnvironments.Staging
        except (KeyError, ValueError):
            return False

    @staticmethod
    def use_environment_from_config(config: configuration.Configuration):
        if IdentifiersProvider.is_staging_environment_enabled(config.config):
            IdentifiersProvider.use_staging()
        else:
            IdentifiersProvider.use_default()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio

import octobot.community.supabase_backend.supabase_realtime_socket as supabase_realtime_socket
import octobot.community.supabase_backend.supabase_realtime_channel as supabase_realtime_channel


class AuthenticatedSupabaseRealtimeClient:
    """
    Full rewrite of https://github.com/supabase-community/supabase-py/blob/develop/supabase/lib/realtime_client.py
    as it can't be used as is and is too far from the js implementation.

    AuthenticatedSupabaseRealtimeClient tries to expose a similar logic and public API
    as https://github.com/supabase/realtime-js/blob/master/src/RealtimeClient.ts
    """
    AUTH_TIMEOUT = 60

    def __init__(self, realtime_url: str, supabase_key: str):
        self.supabase_key = supabase_key
        self.socket = supabase_realtime_socket.AuthenticatedSupabaseRealtimeSocket(
            realtime_url,
            auto_reconnect=True,
            params={"apikey": supabase_key}
        )
        self.channels = []
        self.access_token = None
        self.update_auth_tasks = []

    async def channel(self, schema: str, table_name: str) \
            -> supabase_realtime_channel.AuthenticatedSupabaseRealtimeChannel:
        await self._ensure_connection()
        chan = self.socket.set_channel(schema, table_name)
        chan.update_auth_payload(self._get_auth_payload_update())
        self.channels.append(chan)
        return chan

    def set_auth(self, access_token, loop=None):
        # similar to https://github.com/supabase/realtime-js/blob/master/src/RealtimeClient.ts#L273
        self.access_token = access_token
        loop = loop or asyncio.get_running_loop()
        if access_token is None:
            # access_token is None: no user is signed in
            if not self.socket.closed:
                # close open realtime clients when user signs out
                self.update_auth_tasks.append(
                    loop.create_task(self.close())
                )
            return
        # access_token is not None: use is authenticated
        channel_auth_update_coros = []
        for channels in self.socket.channels.values():
            for channel in channels:
                channel.update_auth_payload(self._get_auth_payload_update())
                # channel already connected (joined), auth into channel if necessary
                if channel.joined_once and channel.is_joined():
                    channel_auth_update_coros.append(channel.auth())
        # user is signed in: can update channel auth
        self.update_auth_tasks.append(
            loop.create_task(self._on_successful_auth(self.socket.closed, channel_auth_update_coros))
        )

    async def close(self):
        await self.socket.close()
        self.update_auth_tasks.clear()

    async def _on_successful_auth(self, should_reopen, coros):
        if should_reopen and self.socket.channels:
            # connection might have been closed in a previous sign-out
            await self._ensure_connection()
        if coros:
            await asyncio.gather(*coros)

    def _get_auth_payload_update(self):
        return {'access_token': self.access_token}

    async def _ensure_connection(self):
        if (not self.socket.connected) or self.socket.closed:
            await self.socket.aconnect()
            await self.socket.subscribe_channels()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import uuid
import gmqtt
import json
import zlib
import asyncio
import packaging.version as packaging_version

import octobot_commons.enums as commons_enums
import octobot_commons.errors as commons_errors
import octobot_commons.constants as commons_constants
import octobot.community.errors as errors
import octobot.community.feeds.abstract_feed as abstract_feed
import octobot.constants as constants


class CommunityMQTTFeed(abstract_feed.AbstractFeed):
    MQTT_VERSION = gmqtt.constants.MQTTv311
    MQTT_BROKER_PORT = 1883
    RECONNECT_DELAY = 15
    RECONNECT_ENSURE_DELAY = 1
    MAX_MESSAGE_ID_CACHE_SIZE = 100
    MAX_SUBSCRIPTION_ATTEMPTS = 5
    DISABLE_RECONNECT_VALUE = -2
    DEVICE_CREATE_TIMEOUT = 5 * commons_constants.MINUTE_TO_SECONDS
    DEVICE_CREATION_REFRESH_DELAY = 2
    CONNECTION_TIMEOUT = 10

    # Quality of Service level determines the reliability of the data flow between a client and a message broker.
    # The message may be sent in three ways:
    # QoS 0: the message will be received at most once (also known as “fire and forget”).
    # QoS 1: the message will be received at least once.
    # QoS 2: the message will be received exactly once.
    # from https://www.scaleway.com/en/docs/iot/iot-hub/concepts/#quality-of-service-levels-(qos)
    default_QOS = 1

    def __init__(self, feed_url, authenticator):
        super().__init__(feed_url, authenticator)
        self.mqtt_version = self.MQTT_VERSION
        self.mqtt_broker_port = self.MQTT_BROKER_PORT
        self.default_QOS = self.default_QOS
        self.subscribed = False

        self._mqtt_client: gmqtt.Client = None
        self._valid_auth = True
        self._disconnected = True
        self._device_uuid: str = None
        self._subscription_attempts = 0
        self._subscription_topics = set()
        self._reconnect_task = None
        self._connect_task = None
        self._connected_at_least_once = False
        self._processed_messages = []

    async def start(self):
        self.should_stop = False
        self._device_uuid = self.authenticator.user_account.get_selected_bot_device_uuid()
        try:
            await self._connect()
            if self.is_connected():
                self.logger.info("Successful connection request to mqtt device")
            else:
                self.logger.info("Failed to connect to mqtt device")
        except asyncio.TimeoutError as err:
            self.logger.exception(err, True, f"Timeout when connecting to mqtt device: {err}")
        except Exception as err:
            self.logger.exception(err, True, f"Unexpected error when connecting to mqtt device: {err}")

    async def stop(self):
        self.logger.debug("Stopping ...")
        self.should_stop = True
        await self._stop_mqtt_client()
        if self._reconnect_task is not None and not self._reconnect_task.done():
            self._reconnect_task.cancel()
        if self._connect_task is not None and not self._connect_task.done():
            self._connect_task.cancel()
        self._reset()
        self.logger.debug("Stopped")

    async def restart(self):
        try:
            if not self.should_stop:
                await self.stop()
            await self.start()
        except Exception as err:
            self.logger.exception(err, True, f"{err}")

    def is_up_to_date_with_account(self, user_account):
        try:
            return self._device_uuid == user_account.get_selected_bot_device_uuid()
        except errors.NoBotDeviceError:
            return False

    def _reset(self):
        self._connected_at_least_once = False
        self._subscription_attempts = 0
        self._connect_task = None
        self._valid_auth = True
        self._disconnected = True

    async def _stop_mqtt_client(self):
        if self.is_connected():
            await self._mqtt_client.disconnect()

    def is_connected(self):
        return self._mqtt_client is not None and self._mqtt_client.is_connected and not self._disconnected

    def is_connected_to_remote_feed(self):
        return self.subscribed

    def can_connect(self):
        return self._valid_auth

    async def register_feed_callback(self, channel_type, callback, identifier=None):
        self.is_signal_receiver = True
        topic = self._build_topic(channel_type, identifier)
        try:
            self.feed_callbacks[topic].append(callback)
        except KeyError:
            self.feed_callbacks[topic] = [callback]
        if topic not in self._subscription_topics:
            self._subscription_topics.add(topic)
            if self._valid_auth:
                self._subscribe((topic, ))
            else:
                self.logger.error(f"Can't subscribe to {channel_type.name} feed, invalid authentication")

    def remove_device_details(self):
        self._device_uuid = None

    @staticmethod
    def _build_topic(channel_type, identifier):
        return f"{channel_type.value}/{identifier}"

    async def _on_message(self, client, topic, payload, qos, properties):
        try:
            uncompressed_payload = zlib.decompress(payload).decode()
            self.logger.debug(f"Received message, client_id: {client._client_id}, topic: {topic}, "
                              f"uncompressed payload: {uncompressed_payload}, QOS: {qos}, properties: {properties}")
            parsed_message = json.loads(uncompressed_payload)
        except Exception as err:
            self.logger.exception(err, True, f"Unexpected error when reading message: {err}")
            return
        await self._process_message(topic, parsed_message)

    async def _process_message(self, topic, parsed_message):
        try:
            self._ensure_supported(parsed_message)
            if self._should_process(parsed_message):
                self.update_last_message_time()
                for callback in self._get_callbacks(topic):
                    await callback(parsed_message)
        except commons_errors.UnsupportedError as err:
            self.logger.error(f"Unsupported message: {err}")
        except Exception as err:
            self.logger.exception(err, True, f"Unexpected error when processing message: {err}")

    def _should_process(self, parsed_message):
        if parsed_message[commons_enums.CommunityFeedAttrs.ID.value] in self._processed_messages:
            self.logger.debug(f"Ignored already processed message with id: "
                              f"{parsed_message[commons_enums.CommunityFeedAttrs.ID.value]}")
            return False
        self._processed_messages.append(parsed_message[commons_enums.CommunityFeedAttrs.ID.value])
        if len(self._processed_messages) > self.MAX_MESSAGE_ID_CACHE_SIZE:
            self._processed_messages = [
                message_id
                for message_id in self._processed_messages[self.MAX_MESSAGE_ID_CACHE_SIZE // 2:]
            ]
        return True

    async def send(self, message, channel_type, identifier, **kwargs):
        self.is_signal_emitter = True
        if not self._valid_auth:
            self.logger.warning(f"Can't send {channel_type.name}, invalid feed authentication.")
            return
        topic = self._build_topic(channel_type, identifier)
        self.logger.info(f"Sending message on topic: {topic}, message: {message}")
        self._mqtt_client.publish(
            self._build_topic(channel_type, identifier),
            self._build_message(channel_type, message),
            qos=self.default_QOS
        )

    def _get_callbacks(self, topic):
        for callback in self.feed_callbacks.get(topic, ()):
            yield callback

    def _get_channel_type(self, message):
        return commons_enums.CommunityChannelTypes(message[commons_enums.CommunityFeedAttrs.CHANNEL_TYPE.value])

    def _build_message(self, channel_type, message):
        if message:
            return zlib.compress(
                json.dumps({
                    commons_enums.CommunityFeedAttrs.CHANNEL_TYPE.value: channel_type.value,
                    commons_enums.CommunityFeedAttrs.VERSION.value: constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION,
                    commons_enums.CommunityFeedAttrs.VALUE.value: message,
                    commons_enums.CommunityFeedAttrs.ID.value: str(uuid.uuid4()),  # assign unique id to each message
                }).encode()
            )
        return {}

    def _ensure_supported(self, parsed_message):
        if packaging_version.Version(parsed_message[commons_enums.CommunityFeedAttrs.VERSION.value]) \
                < packaging_version.Version(constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION):
            raise commons_errors.UnsupportedError(
                f"Minimum version: {constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION}"
            )

    def _on_connect(self, client, flags, rc, properties):
        self._disconnected = False
        self.logger.info(f"Connected, client_id: {client._client_id}")
        # There are no subscription when we just connected
        self.subscribed = False
        # Auto subscribe to known topics (mainly used in case of reconnection)
        self._subscribe(self._subscription_topics)

    def _try_reconnect_if_necessary(self, client):
        if self._reconnect_task is None or self._reconnect_task.done():
            self.logger.debug("Trying to reconnect")
            self._reconnect_task = asyncio.create_task(self._reconnect(client))
        else:
            self.logger.debug("A reconnect task is already running")

    async def _reconnect(self, client):
        try:
            try:
                await self._stop_mqtt_client()
            except Exception as e:
                self.logger.debug(f"Ignored error while stopping client: {e}.")
            attempt = 1
            while not self.should_stop:
                delay = 0 if attempt == 1 else self.RECONNECT_DELAY
                error = None
                try:
                    self.logger.info(f"Reconnecting, client_id: {client._client_id} (attempt {attempt})")
                    await self._connect()
                    await asyncio.sleep(self.RECONNECT_ENSURE_DELAY)
                    if self.is_connected():
                        self.logger.info(f"Reconnected, client_id: {client._client_id}")
                        return
                    error = "failed to connect"
                except Exception as err:
                    error = f"{err}"
                finally:
                    self.logger.debug(f"Reconnect attempt {attempt} {'succeeded' if error is None else 'failed'}.")
                    attempt += 1
                self.logger.debug(f"Error while reconnecting: {error}. Trying again in {delay} seconds.")
                await asyncio.sleep(delay)
        finally:
            self.logger.debug("Reconnect task complete")

    def _on_disconnect(self, client, packet, exc=None):
        self._disconnected = True
        self.subscribed = False
        if self.should_stop:
            self.logger.info(f"Disconnected after stop call")
        else:
            if self._connected_at_least_once:
                self.logger.info(f"Disconnected, client_id: {client._client_id}")
                self._try_reconnect_if_necessary(client)
            else:
                if self._connect_task is not None and not self._connect_task.done():
                    self._connect_task.cancel()

    def _on_subscribe(self, client, mid, qos, properties):
        # from https://github.com/wialon/gmqtt/blob/master/examples/resubscription.py#L28
        # in order to check if all the subscriptions were successful, we should first get all subscriptions with this
        # particular mid (from one subscription request)
        subscriptions = client.get_subscriptions_by_mid(mid)
        for subscription, granted_qos in zip(subscriptions, qos):
            # in case of bad suback code, we can resend  subscription
            if granted_qos >= gmqtt.constants.SubAckReasonCode.UNSPECIFIED_ERROR.value:
                self.logger.warning(f"Retrying subscribe to {[s.topic for s in subscriptions]}, "
                                    f"client_id: {client._client_id}, mid: {mid}, "
                                    f"reason code: {granted_qos}, properties {properties}")
                if self._subscription_attempts < self.MAX_SUBSCRIPTION_ATTEMPTS * len(subscriptions):
                    self._subscription_attempts += 1
                    client.resubscribe(subscription)
                else:
                    self.logger.error(f"Max subscription attempts reached, stopping subscription "
                                      f"to {[s.topic for s in subscriptions]}. Are you subscribing to this "
                                      f"strategy on your OctoBot account ?")
                    self.subscribed = False
                    return
            else:
                self._subscription_attempts = 0
                self.subscribed = True
                self.logger.info(f"Subscribed, client_id: {client._client_id}, mid {mid}, QOS: {granted_qos}, "
                                 f"properties {properties}")

    def _register_callbacks(self, client):
        client.on_connect = self._on_connect
        client.on_message = self._on_message
        client.on_disconnect = self._on_disconnect
        client.on_subscribe = self._on_subscribe

    def _update_client_config(self, client):
        default_config = gmqtt.constants.DEFAULT_CONFIG
        # prevent default auto-reconnect as it loop infinitely on windows long disconnections
        default_config.update({
            'reconnect_retries': self.DISABLE_RECONNECT_VALUE,
        })
        client.set_config(default_config)

    async def _connect(self):
        if self._device_uuid is None:
            self._valid_auth = False
            raise errors.BotError("mqtt device uuid is None, impossible to connect client")
        self._mqtt_client = gmqtt.Client(self.__class__.__name__)
        self._update_client_config(self._mqtt_client)
        self._register_callbacks(self._mqtt_client)
        self._mqtt_client.set_auth_credentials(self._device_uuid, None)
        self.logger.debug(f"Connecting client using device "
                          f"'{self.authenticator.user_account.get_selected_bot_device_name()}'")
        self._connect_task = asyncio.create_task(
            self._mqtt_client.connect(self.feed_url, self.mqtt_broker_port, version=self.MQTT_VERSION)
        )
        try:
            await asyncio.wait_for(self._connect_task, self.CONNECTION_TIMEOUT)
            self._connected_at_least_once = True
        except asyncio.CancelledError:
            # got cancelled by on_disconnect, can't connect
            self.logger.error(f"Can't connect to server, make sure that your device uuid is valid. "
                              f"Current mqtt uuid is: {self._device_uuid}")
            self._valid_auth = False
        except asyncio.TimeoutError as err:
            message = "Timeout error when trying to connect to mqtt device"
            self.logger.debug(message)
            raise asyncio.TimeoutError(message) from err

    def _subscribe(self, topics):
        if not topics:
            self.logger.debug("No topic to subscribe to, skipping subscribe for now")
            return
        subscriptions = [
            gmqtt.Subscription(topic, qos=self.default_QOS)
            for topic in topics
        ]
        self.logger.debug(f"Subscribing to {', '.join(topics)}")
        self._mqtt_client.subscribe(subscriptions)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import random
import time

from config import SIMULATOR_LAST_PRICES_TO_CHECK


async def fill_limit_or_stop_order(limit_or_stop_order, min_price, max_price):
    last_prices = []
    limit_or_stop_order.created_time = time.time()
    for i in range(0, SIMULATOR_LAST_PRICES_TO_CHECK):
        last_prices.insert(i, {})
        last_prices[i]["price"] = random.uniform(min_price, max_price)
        last_prices[i]["timestamp"] = time.time()

    limit_or_stop_order.last_prices = last_prices
    await limit_or_stop_order.update_order_status()


async def fill_market_order(market_order, price):
    last_prices = [{
        "price": price
    }]

    market_order.last_prices = last_prices
    await market_order.update_order_status()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import aiohttp

import octobot_commons.logging


class ErrorsUploader:
    """
    ErrorsUploader manages errors posts to the error url
    """

    def __init__(self, upload_url):
        self.upload_url = upload_url
        self.loop = None
        self.upload_delay = 5

        self._to_upload_errors = []
        self._upload_task = None

        self.logger = octobot_commons.logging.get_logger(self.__class__.__name__)

    def schedule_error_upload(self, error):
        """
        Called to schedule an error upload
        :param error: the octobot_commons.logging.error_model.Error to upload
        """
        self._add_error(error)
        self._ensure_upload_task()

    def _add_error(self, error):
        for existing_error in self._to_upload_errors:
            # first check if error is equivalent to an existing one
            if existing_error.is_equivalent(error):
                existing_error.merge_equivalent(error)
                return
        self._to_upload_errors.append(error)

    def _ensure_upload_task(self):
        try:
            if self._ensure_event_loop() and (self._upload_task is None or self._upload_task.done()):
                self._schedule_upload()
        except Exception as err:
            self.logger.exception(
                err,
                True,
                f"Error when uploading exception: {err}",
                skip_post_callback=True,
            )

    async def _upload_errors(self, session, errors):
        async with session.post(self.upload_url, json=self._get_formatted_errors(errors)) as resp:
            if resp.status != 200:
                self.logger.error(
                    f"Impossible to upload error : status code: {resp.status}, text: {await resp.text()}",
                    skip_post_callback=True
                )

    @staticmethod
    def _get_formatted_errors(errors):
        return [error.to_dict() for error in errors]

    def _schedule_upload(self):
        self._upload_task = self.loop.create_task(
            self._upload_soon()
        )

    async def _upload_soon(self):
        try:
            await asyncio.sleep(self.upload_delay)

            if self._to_upload_errors:
                async with aiohttp.ClientSession() as session:
                    errors = self._to_upload_errors
                    self._to_upload_errors = []
                    await self._upload_errors(session, errors)
                    self.logger.debug(f"Uploaded {len(errors)} errors")
        except Exception as err:
            self.logger.exception(
                err, True, f"Error when uploading exception: {err}", skip_post_callback=True
            )
        finally:
            if self._to_upload_errors:
                # reschedule if new errors arrived during upload
                self._schedule_upload()

    def _ensure_event_loop(self):
        if self.loop is not None:
            if self.loop.is_running():
                return True
            # otherwise, use the current loop
        try:
            self.loop = asyncio.get_event_loop()
            return True
        except RuntimeError:
            return False



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import os
import pathlib
import dotenv
import decimal

import octobot_commons.os_util as os_util
import octobot_commons.enums
import octobot.enums

# make constants visible
from octobot import (
    PROJECT_NAME,
    AUTHOR,
    VERSION,
    LONG_VERSION,
)

# load environment variables from .env file if exists
DOTENV_PATH = os.getenv("DOTENV_PATH", os.path.curdir)
dotenv.load_dotenv(os.path.join(DOTENV_PATH, ".env"), verbose=False)

# OctoBot urls
OCTOBOT_WEBSITE_URL = os.getenv("OCTOBOT_ONLINE_URL", "https://www.octobot.cloud")
OCTOBOT_DOCS_URL = os.getenv("DOCS_OCTOBOT_ONLINE_URL", "https://www.octobot.cloud/guides")
EXCHANGES_DOCS_URL = os.getenv("DOCS_OCTOBOT_ONLINE_URL", "https://www.octobot.cloud/guides/exchanges/")
DEVELOPER_DOCS_URL = os.getenv("DOCS_OCTOBOT_ONLINE_URL", "https://www.octobot.cloud/guides/developers/")
STATIC_OCTOBOT_ONLINE = os.getenv("TENTACLES_OCTOBOT_ONLINE_URL", "https://static.octobot.online")
OCTOBOT_FEEDBACK = os.getenv("FEEDBACK_OCTOBOT_ONLINE_URL", "https://feedback.octobot.cloud/")
TENTACLES_REPOSITORY = "tentacles"
BETA_TENTACLES_REPOSITORY = "dev-tentacles"
OFFICIALS = "officials"
TENTACLE_CATEGORY = "full"
TENTACLE_PACKAGE_NAME = "base"
BETA_TENTACLE_PACKAGE_NAME = "beta"
TENTACLE_PACKAGES = "packages"
COMPILED_TENTACLE_CATEGORY = "extra"

OCTOBOT_DONATION_URL = "https://forms.gle/Bagagc7dyjJGDT1t9"
OCTOBOT_FEEDBACK_FORM_URL = "https://goo.gl/forms/vspraniXPY7rvtKN2"
OCTOBOT_BETA_PROGRAM_FORM_URL = "https://octobot.click/docs-join-beta"
AUTOMATION_FEEDBACK_FORM_ID = "n9NKMV"
WELCOME_FEEDBACK_FORM_ID = os.getenv("WELCOME_FEEDBACK_FORM_ID", None)

COMMUNITY_FEED_CURRENT_MINIMUM_VERSION = "1.0.0"
COMMUNITY_FEED_DEFAULT_TYPE = octobot.enums.CommunityFeedType.SupabaseFeed

# production env SHOULD ONLY BE USED THROUGH CommunityIdentifiersProvider
OCTOBOT_COMMUNITY_LANDING_URL = os.getenv("COMMUNITY_SERVER_URL", "https://octobot.cloud")
OCTOBOT_COMMUNITY_URL = os.getenv("COMMUNITY_SERVER_URL", "https://app.octobot.cloud")
OCTOBOT_COMMUNITY_RECOVER_PASSWORD_URL = OCTOBOT_COMMUNITY_URL
# default env
COMMUNITY_BACKEND_URL = os.getenv("COMMUNITY_BACKEND_URL", "https://nwhpvrguwcihhizrnyoe.supabase.co")
COMMUNITY_BACKEND_KEY = os.getenv("COMMUNITY_BACKEND_KEY", "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6Im53aHB2cmd1d2NpaGhpenJueW9lIiwicm9sZSI6ImFub24iLCJpYXQiOjE2OTU2NDQxMDcsImV4cCI6MjAxMTIyMDEwN30.AILcgv0l6hl_0IUEPlWh1wiu9RIpgrkGZGERM5uXftE")

# staging env SHOULD ONLY BE USED THROUGH CommunityIdentifiersProvider
STAGING_OCTOBOT_COMMUNITY_LANDING_URL = os.getenv("COMMUNITY_SERVER_URL", "https://beta.octobot.cloud")
STAGING_OCTOBOT_COMMUNITY_URL = os.getenv("COMMUNITY_SERVER_URL", "https://app-beta.octobot.cloud/")
STAGING_COMMUNITY_RECOVER_PASSWORD_URL = STAGING_OCTOBOT_COMMUNITY_URL
STAGING_COMMUNITY_BACKEND_URL = os.getenv("COMMUNITY_BACKEND_URL", "https://wmfkgvgzokyzhvxowbyg.supabase.co")
STAGING_COMMUNITY_BACKEND_KEY = os.getenv("COMMUNITY_BACKEND_KEY", "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6IndtZmtndmd6b2t5emh2eG93YnlnIiwicm9sZSI6ImFub24iLCJpYXQiOjE2OTE0NDA1MTEsImV4cCI6MjAwNzAxNjUxMX0.YZQl7LYgvnzO_Jizs0UKfPEaqPoV2EwhjunH8gime8o")

# production env, ignored by CommunityIdentifiersProvider
COMMUNITY_PRODUCTION_BACKEND_URL = os.getenv("COMMUNITY_PRODUCTION_BACKEND_URL", COMMUNITY_BACKEND_URL)
COMMUNITY_PRODUCTION_BACKEND_KEY = os.getenv("COMMUNITY_PRODUCTION_BACKEND_KEY", COMMUNITY_BACKEND_KEY)

CONFIG_COMMUNITY = "community"
CONFIG_COMMUNITY_BOT_ID = "bot_id"
CONFIG_COMMUNITY_ENVIRONMENT = "environment"
USE_BETA_EARLY_ACCESS = os_util.parse_boolean_environment_var("USE_BETA_EARLY_ACCESS", "false")
USER_ACCOUNT_EMAIL = os.getenv("USER_ACCOUNT_EMAIL", "")
USER_PASSWORD_TOKEN = os.getenv("USER_PASSWORD_TOKEN", None)
COMMUNITY_BOT_ID = os.getenv("COMMUNITY_BOT_ID", "")
IS_DEMO = os_util.parse_boolean_environment_var("IS_DEMO", "False")
IS_CLOUD_ENV = os_util.parse_boolean_environment_var("IS_CLOUD_ENV", "false")
CAN_INSTALL_TENTACLES = os_util.parse_boolean_environment_var("CAN_INSTALL_TENTACLES", str(not IS_CLOUD_ENV))
TRACKING_ID = os.getenv("TRACKING_ID", "eoe1stwyun" if IS_DEMO else "eoe06soct7" if IS_CLOUD_ENV else "f726lk9q59")
# Profiles download urls to import at startup if missing, split by ","
TO_DOWNLOAD_PROFILES = os.getenv("TO_DOWNLOAD_PROFILES", None)
# Profiles to force select at startup, identified by profile id, download url or name
FORCED_PROFILE = os.getenv("FORCED_PROFILE", None)
RUN_IN_MAIN_THREAD = os.getenv("RUN_IN_MAIN_THREAD", False)

OCTOBOT_BINARY_PROJECT_NAME = "OctoBot-Binary"

# limits
UNLIMITED_ALLOWED = -1
MAX_ALLOWED_EXCHANGES = int(os.getenv("MAX_ALLOWED_EXCHANGES", str(UNLIMITED_ALLOWED)))
MAX_ALLOWED_SYMBOLS = int(os.getenv("MAX_ALLOWED_SYMBOLS", str(UNLIMITED_ALLOWED)))
MAX_ALLOWED_TIME_FRAMES = int(os.getenv("MAX_ALLOWED_TIME_FRAMES", str(UNLIMITED_ALLOWED)))
MAX_ALLOWED_BACKTESTING_CANDLES_HISTORY = int(os.getenv("MAX_ALLOWED_BACKTESTING_CANDLES_HISTORY",
                                                        str(UNLIMITED_ALLOWED)))
ENABLE_AUTOMATIONS = os_util.parse_boolean_environment_var("ENABLE_AUTOMATIONS", "True")
ENABLE_BACKTESTING = os_util.parse_boolean_environment_var("ENABLE_BACKTESTING", "True")
ENABLE_ADVANCED_INTERFACE = os_util.parse_boolean_environment_var("ENABLE_ADVANCED_INTERFACE", "True")
ENABLE_STRATEGY_OPTIMIZER = os_util.parse_boolean_environment_var("ENABLE_STRATEGY_OPTIMIZER", "True")

# tentacles
ENV_TENTACLES_URL = "TENTACLES_URL"
ENV_COMPILED_TENTACLES_URL = "COMPILED_TENTACLES_URL"
ENV_TENTACLES_REPOSITORY = "TENTACLES_REPOSITORY"
ENV_BETA_TENTACLES_REPOSITORY = "BETA_TENTACLES_REPOSITORY"
ENV_TENTACLES_URL_TAG = "TENTACLES_URL_TAG"
ENV_TENTACLE_PACKAGE_NAME = "TENTACLE_PACKAGE_NAME"
ENV_BETA_TENTACLES_PACKAGE_NAME = "BETA_TENTACLES_PACKAGE_NAME"
ENV_TENTACLES_PACKAGES_TYPE = "TENTACLES_PACKAGES_TYPE"
ENV_TENTACLES_PACKAGES_SOURCE = "TENTACLES_PACKAGES_SOURCE"
ENV_COMPILED_TENTACLES_CATEGORY = "COMPILED_TENTACLES_CATEGORY"
ENV_COMPILED_TENTACLES_PACKAGES_TYPE = "COMPILED_TENTACLES_PACKAGES_TYPE"
ENV_TENTACLE_CATEGORY = "TENTACLE_CATEGORY"
ENV_COMPILED_TENTACLES_SUBCATEGORY = "COMPILED_TENTACLES_SUBCATEGORY"
TENTACLES_REQUIRED_VERSION = f"{os.getenv(ENV_TENTACLES_URL_TAG, LONG_VERSION)}"
# VERSION_PLACEHOLDER will be replaced by the current OctoBot version in ADDITIONAL_TENTACLES_PACKAGE_URL
# ADDITIONAL_TENTACLES_PACKAGE_URL can contain multiple urls separated by a ","
ADDITIONAL_TENTACLES_PACKAGE_URL = os.getenv("ADDITIONAL_TENTACLES_PACKAGE_URL", None)
# url example: 	https://plop.fr/tentacles/123/latest/any_platform.zip
# url example: 	https://plop.fr/tentacles/.../VERSION_PLACEHOLDER/any_platform.zip,https://plop.fr/any_platform.zip
VERSION_PLACEHOLDER = "VERSION_PLACEHOLDER"
URL_SEPARATOR = ","

DEFAULT_TENTACLES_PACKAGE_NAME = "OctoBot-Default-Tentacles"

# logs
LOGS_FOLDER = "logs"
ENV_TRADING_ENABLE_DEBUG_LOGS = os_util.parse_boolean_environment_var("ENV_TRADING_ENABLE_DEBUG_LOGS", "False")

# system
ENABLE_CLOCK_SYNCH = os_util.parse_boolean_environment_var("ENABLE_CLOCK_SYNCH", "True")
ENABLE_SYSTEM_WATCHER = os_util.parse_boolean_environment_var("ENABLE_SYSTEM_WATCHER", "True")
DUMP_USED_RESOURCES = os_util.parse_boolean_environment_var("DUMP_USED_RESOURCES", "False")
USED_RESOURCES_OUTPUT = os.getenv("USED_RESOURCES_OUTPUT", "system_resources.csv")

# errors
ERRORS_URL = os.getenv("ERRORS_OCTOBOT_ONLINE_URL", "https://errors.octobot.online/")
ERRORS_POST_ENDPOINT = f"{ERRORS_URL}errors"
UPLOAD_ERRORS = os_util.parse_boolean_environment_var("UPLOAD_ERRORS", "False")
DEFAULT_METRICS_ID = "UNSET"

# config types keys
CONFIG_KEY = "config"
TENTACLES_SETUP_CONFIG_KEY = "tentacles_setup"

# terms of service
CONFIG_ACCEPTED_TERMS = "accepted_terms"

# DEBUG
CONFIG_DEBUG_OPTION = "DEV-MODE"
FORCE_ASYNCIO_DEBUG_OPTION = False
EXIT_BEFORE_TENTACLES_AUTO_REINSTALL = os_util.parse_boolean_environment_var("EXIT_BEFORE_TENTACLES_AUTO_REINSTALL", "False")

# Files
# Store the path of the octobot directory from this file since it can change depending on the installation path
# (local sources, python site-packages, ...)
OCTOBOT_FOLDER = pathlib.Path(__file__).parent.absolute()
CONFIG_FOLDER = f"{OCTOBOT_FOLDER}/config"
SCHEMA = "schema"
CONFIG_FILE_SCHEMA = f"{CONFIG_FOLDER}/config_{SCHEMA}.json"
PROFILE_FILE_SCHEMA = f"{CONFIG_FOLDER}/profile_{SCHEMA}.json"
DEFAULT_CONFIG_FILE = f"{CONFIG_FOLDER}/default_config.json"
DEFAULT_PROFILE_FILE = f"{CONFIG_FOLDER}/default_profile.json"
DEFAULT_PROFILE_AVATAR_FILE_NAME = "default_profile.png"
DEFAULT_PROFILE_AVATAR = f"{CONFIG_FOLDER}/{DEFAULT_PROFILE_AVATAR_FILE_NAME}"
LOGGING_CONFIG_FILE = f"{CONFIG_FOLDER}/logging_config.ini"
LOG_FILE = f"{LOGS_FOLDER}/{PROJECT_NAME}.log"

# Optimizer
OPTIMIZER_FORCE_ASYNCIO_DEBUG_OPTION = False
OPTIMIZER_DATA_FILES_FOLDER = f"{OCTOBOT_FOLDER}/strategy_optimizer/optimizer_data_files"
OPTIMIZATION_CAMPAIGN_KEY = "optimization_campaign"

OPTIMIZER_RUNS_FOLDER = "optimizer"
OPTIMIZER_DEFAULT_RANDOMLY_CHOSE_RUNS = True
OPTIMIZER_DEFAULT_REQUIRED_IDLE_CORES = 0
OPTIMIZER_DEFAULT_NOTIFY_WHEN_COMPLETE = False
OPTIMIZER_DEFAULT_QUEUE_SIZE = 10000
OPTIMIZER_DEFAULT_MAX_OPTIMIZER_RUNS = 100000
OPTIMIZER_DEFAULT_GENERATIONS_COUNT = 10
OPTIMIZER_DEFAULT_INITIAL_GENERATION_COUNT = OPTIMIZER_DEFAULT_GENERATIONS_COUNT
OPTIMIZER_DEFAULT_RUN_PER_GENERATION = 80
OPTIMIZER_DEFAULT_MUTATION_PERCENT = 20
OPTIMIZER_DEFAULT_MAX_MUTATION_PROBABILITY_PERCENT = decimal.Decimal(95)
OPTIMIZER_DEFAULT_MIN_MUTATION_PROBABILITY_PERCENT = decimal.Decimal(10)
OPTIMIZER_DEFAULT_MAX_MUTATION_NUMBER_MULTIPLIER = 3
OPTIMIZER_DEFAULT_DB_UPDATE_PERIOD = 15

# Databases
DEFAULT_MAX_TOTAL_RUN_DATABASES_SIZE = 1000000000   # 1GB
ENABLE_RUN_DATABASE_LIMIT = os_util.parse_boolean_environment_var("ENABLE_RUN_DATABASE_LIMIT", "True")
MAX_TOTAL_RUN_DATABASES_SIZE = int(os.getenv("MAX_TOTAL_RUN_DATABASES_SIZE", DEFAULT_MAX_TOTAL_RUN_DATABASES_SIZE))

# Channel
OCTOBOT_CHANNEL = "OctoBot"

# Initialization
REQUIRED_TOPIC_FOR_DATA_INIT = [
    octobot_commons.enums.InitializationEventExchangeTopics.CANDLES,
    octobot_commons.enums.InitializationEventExchangeTopics.CONTRACTS,
    octobot_commons.enums.InitializationEventExchangeTopics.PRICE,
    octobot_commons.enums.InitializationEventExchangeTopics.BALANCE,
]

OCTOBOT_KEY = b'uVEw_JJe7uiXepaU_DR4T-ThkjZlDn8Pzl8hYPIv7w0='



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


class FitnessParameter:
    NAME_KEY = "name"
    WEIGHT_KEY = "weight"
    IS_RATIO_FROM_MAX_KEY = "is_ratio_from_max"

    def __init__(self, name, weight, is_ratio_from_max):
        self.name = name
        self.weight = weight
        self.is_ratio_from_max = is_ratio_from_max
        self.max_ratio_value = None
        self.min_ratio_value = None

    def get_normalized_value(self, raw_value):
        if self.is_ratio_from_max:
            # use ratio if relevant
            return self._get_value_from_ratio(raw_value) * self.weight
        return raw_value * self._get_parameter_normalizer() * self.weight
    
    def _get_value_from_ratio(self, raw_value):
        return (
            raw_value * self._get_parameter_normalizer() if self.max_ratio_value is None
            else (raw_value - self.min_ratio_value) / (self.max_ratio_value - self.min_ratio_value)
        )

    def _get_parameter_normalizer(self):
        return 0.01 if "%" in self.name else 1

    def update_ratio(self, full_result):
        try:
            if self.max_ratio_value is None or full_result[self.name] > self.max_ratio_value:
                self.max_ratio_value = full_result[self.name]
            if self.min_ratio_value is None or full_result[self.name] < self.min_ratio_value:
                self.min_ratio_value = full_result[self.name]
        except KeyError:
            pass

    @classmethod
    def from_dict(cls, param_dict):
        return cls(
            param_dict[cls.NAME_KEY],
            param_dict[cls.WEIGHT_KEY],
            param_dict[cls.IS_RATIO_FROM_MAX_KEY],
        )



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import packaging.version as packaging_version

import octobot.constants as constants
import octobot.configuration_manager as configuration_manager
import octobot.commands as commands
import octobot_commons.logging as logging


class Updater:
    def __init__(self):
        self.logger = logging.get_logger(self.__class__.__name__)

    async def should_be_updated(self):
        """
        :return: True if the updater version is greater than current bot version
        """
        try:
            latest_version = await self.get_latest_version()
            if latest_version is None:
                return False
            return packaging_version.parse(latest_version) > packaging_version.parse(constants.VERSION)
        except TypeError as e:
            self.logger.debug(f"Error when comparing latest with current OctoBot version: {e}")

    async def get_latest_version(self):
        raise NotImplementedError("get_latest_version is not implemented")

    async def update_impl(self) -> bool:
        raise NotImplementedError("update_impl is not implemented")

    async def update_tentacles(self):
        await commands.install_all_tentacles(
            tentacles_url=configuration_manager.get_default_tentacles_url(version=await self.get_latest_version()))

    async def post_update(self):
        await self.update_tentacles()
        commands.restart_bot()

    async def update(self):
        """
        Call updater update_impl and updates tentacles on update success
        """
        if await self.update_impl():
            await self.post_update()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import pytest
import pytest_asyncio
import mock
import asyncio
import json
import time
import websockets

import octobot.community as community
import octobot.constants as constants
import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.enums as commons_enums
import octobot_commons.signals as commons_signals

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

HOST = "127.0.0.1"
PORT = 8765
TOKEN = "acb1"

DATA_DICT = {
    "data": "yes",
    "content": {
        "hi": 1,
        "a": [
            "@@@",
            1
        ]
    }
}


class MockedResponse:
    def __init__(self, status_code=200, json=None):
        self.status_code = status_code
        self.json_resp = json

    def json(self):
        return self.json_resp


def _mocked_signal():
    return commons_signals.Signal(
        "identifier", 
        {
            "key1": 11,
            "key2": "random value"
        }
    )


def _build_message(value):
    return {
        commons_enums.CommunityFeedAttrs.VERSION.value: constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION,
        commons_enums.CommunityFeedAttrs.CHANNEL_TYPE.value: commons_enums.CommunityChannelTypes.SIGNAL.value,
        commons_enums.CommunityFeedAttrs.STREAM_ID.value: 1,
        commons_enums.CommunityFeedAttrs.VALUE.value: value
    }


async def echo_or_signal_reply_handler(websocket, _):
    async for message in websocket:
        parsed_message = json.loads(message)
        if parsed_message.get("command") == "subscribe":
            await websocket.send(json.dumps({"type": "confirm_subscription"}))
        else:
            value = json.loads(json.loads(parsed_message.get("data", "")).get("value"))
            base_message = _build_message(_mocked_signal().to_dict() if value == "signal" else value)
            await websocket.send(json.dumps({"message": base_message}))


@pytest_asyncio.fixture
async def community_echo_server():
    async with websockets.serve(echo_or_signal_reply_handler, HOST, PORT):
        yield


@pytest_asyncio.fixture
async def authenticator():
    community.IdentifiersProvider.use_production()
    auth = community.CommunityAuthentication(None, None)
    auth.user_account._profile_raw_data = {"1": 1}
    auth.is_logged_in = mock.Mock(return_value=True)
    auth.get_backend_headers = mock.Mock(return_value={})
    return auth


@pytest_asyncio.fixture
async def community_mocked_consumer_server():
    mock_consumer = mock.AsyncMock()

    async def mock_handler(websocket, path):
        async for message in websocket:
            await mock_consumer(message)
            parsed_message = json.loads(message)
            if parsed_message.get("command") == "subscribe":
                await websocket.send(json.dumps({"type": "confirm_subscription"}))
            else:
                await websocket.send(json.dumps({"message": _build_message(mock_consumer.call_count)}))
    async with websockets.serve(mock_handler, HOST, PORT):
        yield mock_consumer


@pytest_asyncio.fixture
async def connected_community_feed(authenticator):
    feed = None
    try:
        feed = community.CommunityWSFeed(f"ws://{HOST}:{PORT}", authenticator)
        feed.INIT_TIMEOUT = 1
        with mock.patch.object(feed, "_fetch_stream_identifier", mock.AsyncMock(return_value=1)) \
                as _fetch_stream_identifier_mock:
            await feed.register_feed_callback(commons_enums.CommunityChannelTypes.SIGNAL, mock.AsyncMock())
            _fetch_stream_identifier_mock.assert_called_once_with(None)
            await feed.start()
            yield feed
    finally:
        if feed is not None:
            await feed.stop()


async def test_consume_base_message(community_echo_server, connected_community_feed):
    consume_mock = connected_community_feed.feed_callbacks[commons_enums.CommunityChannelTypes.SIGNAL][None][0]
    consume_mock.assert_not_called()
    await connected_community_feed.send("hiiiii", commons_enums.CommunityChannelTypes.SIGNAL, None)
    await _wait_for_receive()
    consume_mock.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': 'hiiiii'})
    consume_mock.reset_mock()

    await connected_community_feed.send(json.dumps(DATA_DICT), commons_enums.CommunityChannelTypes.SIGNAL, None)
    await _wait_for_receive()
    consume_mock.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': json.dumps(DATA_DICT)})


async def test_consume_signal_message(community_echo_server, connected_community_feed):
    # use strat1 as identifier
    consume_mock = connected_community_feed.feed_callbacks[commons_enums.CommunityChannelTypes.SIGNAL][None][0]
    connected_community_feed.feed_callbacks = {
        commons_enums.CommunityChannelTypes.SIGNAL: {
            "strat1": [consume_mock]
        }
    }
    connected_community_feed._identifier_by_stream_id[1] = "strat1"
    consume_mock.assert_not_called()
    await connected_community_feed.send("signal", commons_enums.CommunityChannelTypes.SIGNAL, "strat1")
    await _wait_for_receive()
    consume_mock.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': _mocked_signal().to_dict()})


async def test_send_base_message(community_mocked_consumer_server, connected_community_feed):
    community_mocked_consumer_server.assert_called_once_with(
        '{"command": "subscribe", "identifier": "{\\"channel\\": \\"Spree::MessageChannel\\"}", "data": {}}'
    )
    community_mocked_consumer_server.reset_mock()
    consume_mock = connected_community_feed.feed_callbacks[commons_enums.CommunityChannelTypes.SIGNAL][None][0]
    consume_mock.assert_not_called()
    await connected_community_feed.send("signal", commons_enums.CommunityChannelTypes.SIGNAL, None)
    await _wait_for_receive()
    community_mocked_consumer_server.assert_called_once_with(
        '{"command": "message", "identifier": "{\\"channel\\": \\"Spree::MessageChannel\\"}", "data": '
        '"{\\"topic\\": \\"t\\", \\"feed_id\\": 1, \\"version\\": \\"1.0.0\\", '
        '\\"value\\": \\"\\\\\\"signal\\\\\\"\\"}"}'
    )
    consume_mock.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': 1})
    consume_mock.reset_mock()

    await connected_community_feed.send(DATA_DICT, commons_enums.CommunityChannelTypes.SIGNAL, None)
    await _wait_for_receive()
    assert community_mocked_consumer_server.call_count == 2
    identifier_str = json.dumps({"channel": "Spree::MessageChannel"})
    sub_data_str = json.dumps({"data": "yes", "content": {"hi": 1, "a": ["@@@", 1]}})
    data_str = json.dumps({"topic": "t", "feed_id": 1, "version": "1.0.0", "value": sub_data_str})
    assert community_mocked_consumer_server.mock_calls[1].args == \
       (json.dumps({"command": "message", "identifier": identifier_str, "data": data_str}), )
    consume_mock.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': 2})


async def test_send_signal_message(community_mocked_consumer_server, connected_community_feed):
    community_mocked_consumer_server.assert_called_once_with(
        '{"command": "subscribe", "identifier": "{\\"channel\\": \\"Spree::MessageChannel\\"}", "data": {}}'
    )
    community_mocked_consumer_server.reset_mock()
    consume_mock = connected_community_feed.feed_callbacks[commons_enums.CommunityChannelTypes.SIGNAL][None][0]
    consume_mock.assert_not_called()
    await connected_community_feed.send(_mocked_signal().to_dict(),
                                        commons_enums.CommunityChannelTypes.SIGNAL, None)
    await _wait_for_receive()

    identifier_str = json.dumps({"channel": "Spree::MessageChannel"})
    sub_data_str = json.dumps(_mocked_signal().to_dict())
    data_str = json.dumps({"topic": "t", "feed_id": 1, "version": "1.0.0", "value": sub_data_str})
    community_mocked_consumer_server.assert_called_once_with(
        json.dumps({"command": "message", "identifier": identifier_str, "data": data_str}))
    consume_mock.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': 1})


async def test_reconnect(authenticator):
    server = client = None
    try:
        server = await websockets.serve(echo_or_signal_reply_handler, HOST, PORT)
        client_handler = mock.AsyncMock()
        client = community.CommunityWSFeed(f"ws://{HOST}:{PORT}", authenticator)
        client.RECONNECT_DELAY = 0
        await client.register_feed_callback(commons_enums.CommunityChannelTypes.SIGNAL, client_handler)
        await client.start()

        # 1. ensure client is both receiving and sending messages
        client_handler.assert_not_called()
        await client.send("plop", commons_enums.CommunityChannelTypes.SIGNAL, None)
        await _wait_for_receive()
        client_handler.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': "plop"})
        client_handler.reset_mock()
        await client.send("hii", commons_enums.CommunityChannelTypes.SIGNAL, None)
        await _wait_for_receive()
        client_handler.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': "hii"})
        client_handler.reset_mock()

        # 2. client is disconnected (server closes)
        server.close()
        await server.wait_closed()
        with pytest.raises(websockets.ConnectionClosed):
            await client.send("hii", commons_enums.CommunityChannelTypes.SIGNAL, None,
                              reconnect_if_necessary=False)
        await _wait_for_receive()
        client_handler.assert_not_called()

        # 3. client is reconnected (server restarts)
        server = await websockets.serve(echo_or_signal_reply_handler, HOST, PORT)
        # ensure the previous message was not get sent upon reconnection
        client_handler.assert_not_called()

        # 4. re-exchange message using the same client (reconnected through send method)
        assert not client.is_connected()
        await client.send("plop", commons_enums.CommunityChannelTypes.SIGNAL, None)
        await _wait_for_connection_and_subscribe(client)
        await _wait_for_receive()
        client_handler.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': "plop"})
        client_handler.reset_mock()
        await client.send("hii", commons_enums.CommunityChannelTypes.SIGNAL, None)
        await _wait_for_receive()
        client_handler.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': "hii"})
        client_handler.reset_mock()

        # 5. re disconnect server
        server.close()
        await server.wait_closed()
        with pytest.raises(websockets.ConnectionClosed):
            await client.send("hii", commons_enums.CommunityChannelTypes.SIGNAL, None,
                              reconnect_if_necessary=False)
        await _wait_for_receive()
        client_handler.assert_not_called()

        # 6. wait for client reconnection, receive signal as soon as possible
        server = await websockets.serve(echo_or_signal_reply_handler, HOST, PORT)
        client_handler.assert_not_called()
        # wait for client reconnection
        await _wait_for_connection_and_subscribe(client)

        # 7. send message from server with a consuming only client
        await next(iter(server.websockets)).send(json.dumps({"message": _build_message("greetings")}))
        await _wait_for_receive()
        assert client.is_connected()
        # client_handler is called as the consume task did reconnect the client
        client_handler.assert_called_once_with({'v': '1.0.0', 't': 't', 'i': 1, 's': "greetings"})

    finally:
        if client is not None:
            await client.stop()
        if server is not None:
            server.close()
            await server.wait_closed()


async def _wait_for_connection_and_subscribe(client):
    t0 = time.time()
    while time.time() - t0 < 5:
        if client.is_connected() and client.subscribed:
            break
        else:
            await asyncio.sleep(0.05)
    assert client.is_connected()
    assert client.subscribed


async def _wait_for_receive(wait_cycles=8):
    # 5 necessary wait_cycles for both sending, receiving, replying and receiving a reply
    for _ in range(wait_cycles):
        # wait for websockets lib trigger client
        await asyncio_tools.wait_asyncio_next_cycle()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio

import octobot_commons.logging as logging
import octobot_commons.enums as common_enums
import octobot_commons.tentacles_management as tentacles_management
import octobot_tentacles_manager.api as tentacles_manager_api
import octobot.automation.bases.abstract_trigger_event as abstract_trigger_event
import octobot.automation.bases.abstract_condition as abstract_condition
import octobot.automation.bases.abstract_action as abstract_action
import octobot.constants as constants
import octobot.errors as errors


class AutomationDetails:
    def __init__(self, trigger_event, conditions, actions):
        self.trigger_event = trigger_event
        self.conditions = conditions
        self.actions = actions

    def __str__(self):
        return f"Automation with {self.trigger_event.get_name()} trigger, " \
               f"{' ,'.join([condition.get_name() for condition in self.conditions])} conditions and " \
               f"{' ,'.join([action.get_name() for action in self.actions])} actions"


class Automation(tentacles_management.AbstractTentacle):
    USER_INPUT_TENTACLE_TYPE = common_enums.UserInputTentacleTypes.AUTOMATION
    AUTOMATION = "automation"
    AUTOMATIONS = "automations"
    AUTOMATIONS_COUNT = "automations_count"
    TRIGGER_EVENT = "trigger_event"
    CONDITIONS = "conditions"
    ACTIONS = "actions"

    def __init__(self, bot_id, tentacles_setup_config, automations_config=None):
        super().__init__()
        self.logger = logging.get_logger(self.get_name())
        self.bot_id = bot_id
        self.tentacles_setup_config = tentacles_setup_config
        self.automations_config = automations_config
        self.automation_tasks = []
        self.automation_details = []

    def get_local_config(self):
        return self.automations_config

    async def initialize(self) -> None:
        """
        Triggers producers and consumers creation
        """
        if constants.ENABLE_AUTOMATIONS:
            await self.restart()
        else:
            self.logger.info("Automations are disabled")

    @classmethod
    async def get_raw_config_and_user_inputs(
            cls, config, tentacles_setup_config, bot_id
    ):
        tentacle_config = tentacles_manager_api.get_tentacle_config(tentacles_setup_config, cls)
        local_instance = cls.create_local_instance(
            config, tentacles_setup_config, tentacle_config
        )
        user_inputs = {}
        local_instance.init_user_inputs(user_inputs)
        return tentacle_config, list(user_input.to_dict() for user_input in user_inputs.values())

    async def restart(self):
        if not constants.ENABLE_AUTOMATIONS:
            raise errors.DisabledError("Automations are disabled")
        await self.stop()
        self.automations_config = tentacles_manager_api.get_tentacle_config(self.tentacles_setup_config,
                                                                            self.__class__)
        await self.load_and_save_user_inputs(self.bot_id)
        await self.start()

    async def start(self):
        self._create_automation_details()
        self.automation_tasks = [
            asyncio.create_task(self._run_automation(automation_detail))
            for automation_detail in self.automation_details
        ]
        if not self.automation_details:
            self.logger.debug("No automation to start")

    async def stop(self):
        if self.automation_tasks:
            self.logger.debug("Stopping automation tasks")
            for task in self.automation_tasks:
                if not task.done():
                    task.cancel()

    def reset_config(self):
        tentacles_manager_api.update_tentacle_config(
            self.tentacles_setup_config,
            self.__class__,
            {
                self.AUTOMATIONS_COUNT: 0,
                self.AUTOMATIONS: {}
            }
        )

    @classmethod
    def create_local_instance(cls, config, tentacles_setup_config, tentacle_config):
        return cls(None, tentacles_setup_config, automations_config=tentacle_config)

    def _all_possible_steps(self, base_step):
        return tentacles_management.get_all_classes_from_parent(base_step)

    def get_all_steps(self):
        all_events = {
            event.get_name(): event
            for event in self._all_possible_steps(abstract_trigger_event.AbstractTriggerEvent)
        }
        all_conditions = {
            condition.get_name(): condition
            for condition in self._all_possible_steps(abstract_condition.AbstractCondition)
        }
        all_actions = {
            action.get_name(): action
            for action in self._all_possible_steps(abstract_action.AbstractAction)
        }
        return all_events, all_conditions, all_actions

    def _get_default_steps(self):
        import tentacles.Automation.trigger_events as trigger_events_impl
        import tentacles.Automation.conditions as conditions_impl
        import tentacles.Automation.actions as actions_impl
        return trigger_events_impl.PeriodicCheck.get_name(), \
               [conditions_impl.NoCondition.get_name()], \
               [actions_impl.SendNotification.get_name()]

    def init_user_inputs(self, inputs: dict) -> None:
        """
        Called right before starting the tentacle, should define all the tentacle's user inputs unless
        those are defined somewhere else.
        """
        self.automation_details = []
        all_events, all_conditions, all_actions = self.get_all_steps()
        automations_count = self.UI.user_input(self.AUTOMATIONS_COUNT, common_enums.UserInputTypes.INT,
                                               self.automations_config.get(self.AUTOMATIONS_COUNT, 0), inputs,
                                               min_val=0,
                                               title="Number of automations.")
        if not automations_count:
            return
        automations = self.UI.user_input(self.AUTOMATIONS, common_enums.UserInputTypes.OBJECT,
                                         self.automations_config.get(self.AUTOMATIONS, {}), inputs,
                                         title="Automations")
        default_event, default_conditions, default_actions = self._get_default_steps()
        for index in range(1, automations_count + 1):
            automation_id = f"{index}"
            # register trigger events
            self.UI.user_input(automation_id, common_enums.UserInputTypes.OBJECT,
                               automations.get(automation_id, {}), inputs,
                               parent_input_name=self.AUTOMATIONS,
                               title=f"Automation {index}")
            event = self.UI.user_input(self.TRIGGER_EVENT, common_enums.UserInputTypes.OPTIONS,
                                       default_event, inputs,
                                       options=list(all_events),
                                       parent_input_name=automation_id,
                                       title="The trigger for this automation.")
            if event:
                self._apply_user_inputs([event], all_events, inputs, automation_id)
            # register conditions
            conditions = self.UI.user_input(self.CONDITIONS, common_enums.UserInputTypes.MULTIPLE_OPTIONS,
                                            default_conditions, inputs,
                                            options=list(all_conditions),
                                            parent_input_name=automation_id,
                                            title="Conditions for this automation.")
            self._apply_user_inputs(conditions, all_conditions, inputs, automation_id)
            # register actions
            actions = self.UI.user_input(self.ACTIONS, common_enums.UserInputTypes.MULTIPLE_OPTIONS,
                                         default_actions, inputs,
                                         options=list(all_actions),
                                         parent_input_name=automation_id,
                                         title="Actions for this automation.")
            self._apply_user_inputs(actions, all_actions, inputs, automation_id)

    def _apply_user_inputs(self, step_names, step_classes_by_name: dict, inputs, automation_id):
        for step_name in step_names:
            try:
                self._apply_step_user_inputs(step_name, step_classes_by_name[step_name], inputs, automation_id)
            except KeyError:
                self.logger.error(f"Automation step not found: {step_name} (ignored)")

    def _apply_step_user_inputs(self, step_name, step_class, inputs, automation_id):
        step = step_class()
        user_inputs = step.get_user_inputs(self.UI, inputs, step_name)
        if user_inputs:
            self.UI.user_input(
                step_name, common_enums.UserInputTypes.OBJECT,
                user_inputs, inputs,
                parent_input_name=automation_id,
                array_indexes=[0],
                title=f"{step_name} configuration"
            )

    def _is_valid_automation_config(self, automation_config):
        return automation_config.get(self.TRIGGER_EVENT) is not None

    def _create_automation_details(self):
        all_events, all_conditions, all_actions = self.get_all_steps()
        automations_count = self.automations_config.get(self.AUTOMATIONS_COUNT, 0)
        for automation_id, automation_config in self.automations_config.get(self.AUTOMATIONS, {}).items():
            if int(automation_id) > automations_count:
                return
            if not self._is_valid_automation_config(automation_config):
                continue
            event = self._create_step(automation_config, automation_config[self.TRIGGER_EVENT], all_events)
            conditions = [
                self._create_step(automation_config, condition, all_conditions)
                for condition in automation_config[self.CONDITIONS]
            ]
            actions = [
                self._create_step(automation_config, action, all_actions)
                for action in automation_config[self.ACTIONS]
            ]
            self.automation_details.append(AutomationDetails(event, conditions, actions))

    def _create_step(self, automations_config, step_name, classes):
        step = classes[step_name]()
        step.apply_config(automations_config.get(step_name, {}))
        return step

    async def _run_automation(self, automation_detail):
        self.logger.info(f"Starting {automation_detail} automation")
        async for _ in automation_detail.trigger_event.next_event():
            self.logger.debug(f"{automation_detail.trigger_event.get_name()} event triggered")
            if await self._check_conditions(automation_detail):
                await self._process_actions(automation_detail)

    async def _check_conditions(self, automation_detail):
        for condition in automation_detail.conditions:
            if not await condition.evaluate():
                # not all conditions are valid, skip event
                self.logger.debug(f"{condition.get_name()} is not valid: skipping "
                                  f"{automation_detail.trigger_event.get_name()} event")
                return False
            self.logger.debug(f"All conditions are valid for "
                              f"{automation_detail.trigger_event.get_name()} event trigger")
        return True

    async def _process_actions(self, automation_detail):
        for action in automation_detail.actions:
            try:
                self.logger.debug(f"Running {action.get_name()} after "
                                  f"{automation_detail.trigger_event.get_name()} event")
                await action.process()
            except Exception as err:
                self.logger.exception(err, True, f"Error when running action: {err}")




#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import json
import uuid
import gc
import sys
import asyncio
import time

import octobot_commons.logging as commons_logging
import octobot_commons.configuration as commons_configuration
import octobot_commons.databases as commons_databases
import octobot_commons.constants as commons_constants
import octobot_commons.enums as commons_enums
import octobot_commons.list_util as list_util
import octobot_commons.asyncio_tools as asyncio_tools

import octobot_backtesting.api as backtesting_api
import octobot_backtesting.importers as importers

import octobot_evaluators.api as evaluator_api
import octobot_evaluators.constants as evaluator_constants

import octobot_services.api as service_api

import octobot_trading.exchanges as exchanges
import octobot_trading.exchange_data as exchange_data
import octobot_trading.api as trading_api
import octobot_trading.enums as trading_enums

import octobot.logger as logger
import octobot.storage as storage
import octobot.limits as limits
import octobot.constants as constants
import octobot.errors as errors
import octobot.databases_util as databases_util


class OctoBotBacktesting:
    def __init__(
        self,
        backtesting_config,
        tentacles_setup_config,
        symbols_to_create_exchange_classes,
        backtesting_files,
        run_on_common_part_only,
        start_timestamp=None,
        end_timestamp=None,
        enable_logs=True,
        enable_storage=True,
        run_on_all_available_time_frames=False,
        backtesting_data=None,
        name=None,
        config_by_tentacle=None,
        services_config=None,
    ):
        self.logger = commons_logging.get_logger(self.__class__.__name__)
        self.backtesting_config = backtesting_config
        self.tentacles_setup_config = tentacles_setup_config
        self.config_by_tentacle: dict = config_by_tentacle
        self.bot_id = str(uuid.uuid4())
        self.matrix_id = ""
        self.exchange_manager_ids = []
        self.symbols_to_create_exchange_classes = symbols_to_create_exchange_classes
        self.evaluators = []
        self.service_feeds = []
        self.backtesting_files = backtesting_files
        self.backtesting_data = backtesting_data
        self.name = name
        if self.backtesting_data is not None:
            self.backtesting_files = [
                backtesting_file
                for backtesting_file in self.backtesting_data.data_files
            ]
        self.backtesting = None
        self.run_on_common_part_only = run_on_common_part_only
        self.start_time = None
        self.start_timestamp = start_timestamp
        self.end_timestamp = end_timestamp
        self.enable_logs = enable_logs
        self.exchange_type_by_exchange = {}
        self.futures_contract_type = trading_enums.FutureContractType.LINEAR_PERPETUAL
        self.enable_storage = enable_storage
        self.run_on_all_available_time_frames = run_on_all_available_time_frames
        self._has_started = False
        self.has_fetched_data = False
        self.services_config = services_config

    async def initialize_and_run(self):
        if not constants.ENABLE_BACKTESTING:
            raise errors.DisabledError("Backtesting is disabled")
        self.logger.info(f"Starting on {self.backtesting_files} with {self.symbols_to_create_exchange_classes}")
        self.start_time = time.time()
        await commons_databases.init_bot_storage(
            self.bot_id,
            databases_util.get_run_databases_identifier(
                self.backtesting_config,
                self.tentacles_setup_config,
                enable_storage=self.enable_storage,
            ),
            False
        )
        await self._init_matrix()
        await self._init_backtesting()
        await self._init_evaluators()
        await self._init_service_feeds()
        min_timestamp, max_timestamp = await self._configure_backtesting_time_window()
        await self._init_exchanges()
        self._ensure_limits()
        await self._create_evaluators()
        await self._fetch_backtesting_extra_data_if_any(
            min_timestamp, max_timestamp
        )
        await self._create_service_feeds()
        await backtesting_api.start_backtesting(self.backtesting)
        if logger.BOT_CHANNEL_LOGGER is not None and self.enable_logs:
            await self.start_loggers()
        self._has_started = True

    async def stop_importers(self):
        if self.backtesting is not None:
            # Close databases
            for importer in backtesting_api.get_importers(self.backtesting):
                if importer is not None:
                    await backtesting_api.stop_importer(importer)

    async def stop(self, memory_check=False, should_raise=False):
        symbols_by_exchange = {
            exchange: [str(s) for s in symbols]
            for exchange, symbols in self.symbols_to_create_exchange_classes.items()
        }
        self.logger.debug(f"Stopping for {self.backtesting_files} with {symbols_by_exchange}")
        exchange_managers = []
        try:
            if self.backtesting is None:
                self.logger.warning("No backtesting to stop, there was probably an issue when starting the backtesting")
            else:
                exchange_managers = trading_api.get_exchange_managers_from_exchange_ids(self.exchange_manager_ids)
                if exchange_managers and self.enable_storage and self._has_started:
                    try:
                        for exchange_manager in exchange_managers:
                            await trading_api.store_history_in_run_storage(exchange_manager)
                    except Exception as e:
                        self.logger.exception(e, True, f"Error when saving exchange historical data: {e}")
                    try:
                        await self._store_metadata(exchange_managers)
                        self.logger.info(f"Stored backtesting run metadata")
                    except Exception as e:
                        self.logger.exception(e, True, f"Error when saving run metadata: {e}")
                await backtesting_api.stop_backtesting(self.backtesting)
            try:
                for exchange_manager in exchange_managers:
                    await trading_api.stop_exchange(exchange_manager)
            except KeyError:
                # exchange managers are not added in global exchange list when an exception occurred
                pass
            # close run databases
            await commons_databases.close_bot_storage(self.bot_id)
            # stop evaluators
            for evaluators in self.evaluators:
                # evaluators by type
                for evaluator in evaluators:
                    # evaluator instance
                    if evaluator is not None:
                        await evaluator_api.stop_evaluator(evaluator)
            # close all caches (if caches have to be used later on, they can always be re-opened)
            await commons_databases.CacheManager().close_cache(commons_constants.UNPROVIDED_CACHE_IDENTIFIER)
            try:
                await evaluator_api.stop_all_evaluator_channels(self.matrix_id)
            except KeyError:
                if self.evaluators:
                    raise
                # matrix not created (and no evaluator, KeyError is expected)
            evaluator_api.del_evaluator_channels(self.matrix_id)
            evaluator_api.del_matrix(self.matrix_id)
            for service_feed in self.service_feeds:
                await service_api.stop_service_feed(service_feed)
        except Exception as e:
            self.logger.exception(e, True, f"Error when stopping independent backtesting: {e}")
            if should_raise:
                raise
        finally:
            # call stop_importers in case it has not been called already
            if self.backtesting_data is None:
                await self.stop_importers()
            # let cancelled backtesting tasks complete before returning
            await asyncio_tools.wait_asyncio_next_cycle()
            if memory_check:
                to_reference_check = exchange_managers + [self.backtesting]
                # Call at the next loop iteration to first let coroutines get cancelled
                # (references to coroutine and caller objects are kept while in async loop)
                asyncio.get_event_loop().call_soon(self.memory_leak_checkup, to_reference_check)
            self.backtesting = None

    def memory_leak_checkup(self, to_check_elements):
        gc.collect()
        self.logger.debug(f"Memory leak checking {[e.__class__.__name__ for e in to_check_elements]}")
        memory_leak_errors = []
        for i in range(len(to_check_elements)):
            if sys.getrefcount(to_check_elements[i]) > 2:
                # Using PyCharm debugger, right click on the element variable and use "Find references"
                # Warning: Python debugger can add references when watching an element
                element = to_check_elements[i]
                # Now expect 3 references because the above element variable adds a reference
                memory_leak_errors.append(f" Too many remaining references on the {element.__class__.__name__} element "
                                          f"after {self.__class__.__name__} run, the garbage collector won't free it "
                                          f"(expected a maximum of 3 references): {sys.getrefcount(element)} actual "
                                          f"references ({element})")
        if memory_leak_errors:
            errors = "\n".join(memory_leak_errors)
            raise AssertionError(
                f"[Dev oriented error: no effect on backtesting result, please report if you see it]: {errors}"
            )

    # Use check_remaining_objects to check remaining objects from garbage collector after calling stop().
    # Warning: can take a long time when a lot of objects exist
    def check_remaining_objects(self):
        objects_leak_errors = []
        exchanges_count = len(self.exchange_manager_ids)
        to_watch_objects = (exchange_data.ExchangeSymbolData, exchanges.ExchangeManager,
                            exchanges.ExchangeSimulator, exchange_data.OHLCVUpdaterSimulator)
        objects_references = {obj: (0, []) for obj in to_watch_objects}
        expected_max_objects_references = {
            exchange_data.ExchangeSymbolData: exchanges_count + 1,
            exchanges.ExchangeManager: exchanges_count + 1,
            exchanges.ExchangeSimulator: exchanges_count,
            exchange_data.OHLCVUpdaterSimulator: exchanges_count
        }
        # trigger garbage collector to get a fresh memory picture
        gc.collect()
        for obj in gc.get_objects():
            if isinstance(obj, to_watch_objects):
                if isinstance(obj, exchanges.ExchangeManager) and not obj.is_initialized:
                    # Ignore exchange managers that have not been initialized
                    # and are irrelevant. Pytest fixtures can also retain references failing tests
                    continue
                objects_references[type(obj)][1].append(obj)
                objects_references[type(obj)] = (objects_references[type(obj)][0] + 1,
                                                 objects_references[type(obj)][1])

        for obj, max_ref in expected_max_objects_references.items():
            if objects_references[obj][0] > max_ref:
                objects_leak_errors.append(_get_remaining_object_error(obj,
                                                                       max_ref,
                                                                       objects_references[obj]))

        if objects_leak_errors:
            errors = "\n".join(objects_leak_errors)
            raise AssertionError(
                f"[Dev oriented error: no effect on backtesting result, please report if you see it]: {errors}"
            )

    async def _store_metadata(self, exchange_managers):
        run_db = commons_databases.RunDatabasesProvider.instance().get_run_db(self.bot_id)
        await run_db.flush()
        user_inputs = await commons_configuration.get_user_inputs(run_db)
        await storage.store_run_metadata(
            self.bot_id,
            exchange_managers,
            self.start_time,
            user_inputs=user_inputs,
        )
        metadata = await storage.store_backtesting_run_metadata(
            exchange_managers,
            self.start_time,
            user_inputs,
            commons_databases.RunDatabasesProvider.instance().get_run_databases_identifier(self.bot_id),
            self.name
        )
        self.logger.info(f"Backtesting metadata:\n{json.dumps(metadata, indent=4)}")

    async def _init_matrix(self):
        self.matrix_id = evaluator_api.create_matrix()

    async def _init_evaluators(self):
        await evaluator_api.initialize_evaluators(
            self.backtesting_config, self.tentacles_setup_config, config_by_evaluator=self.config_by_tentacle
        )
        if (not self.backtesting_config[commons_constants.CONFIG_TIME_FRAME]) and \
           evaluator_constants.CONFIG_FORCED_TIME_FRAME in self.backtesting_config:
            self.backtesting_config[commons_constants.CONFIG_TIME_FRAME] = self.backtesting_config[
                evaluator_constants.CONFIG_FORCED_TIME_FRAME
            ]
        await evaluator_api.create_evaluator_channels(self.matrix_id, is_backtesting=True)

    async def _init_service_feeds(self):
        service_feed_factory = service_api.create_service_feed_factory(self.backtesting_config,
                                                                       asyncio.get_event_loop(),
                                                                       self.bot_id)
        self.service_feeds = [service_feed_factory.create_service_feed(feed)
                              for feed in service_feed_factory.get_available_service_feeds(True)]

    def _ensure_limits(self):
        for exchange_id in self.exchange_manager_ids:
            exchange_configuration = trading_api.get_exchange_configuration_from_exchange_id(exchange_id)
            time_frames = exchange_configuration.available_required_time_frames
            symbols = exchange_configuration.symbols
            start_time, end_time = trading_api.get_exchange_backtesting_time_window(
                trading_api.get_exchange_manager_from_exchange_id(exchange_id)
            )
            limits.ensure_backtesting_limits(
                self.symbols_to_create_exchange_classes.keys(), symbols, time_frames, start_time, end_time
            )

    async def _create_evaluators(self):
        for exchange_id in self.exchange_manager_ids:
            exchange_configuration = trading_api.get_exchange_configuration_from_exchange_id(exchange_id)
            self.evaluators = await evaluator_api.create_and_start_all_type_evaluators(
                self.tentacles_setup_config,
                matrix_id=self.matrix_id,
                exchange_name=exchange_configuration.exchange_name,
                bot_id=self.bot_id,
                symbols_by_crypto_currencies=exchange_configuration.symbols_by_crypto_currencies,
                symbols=exchange_configuration.symbols,
                time_frames=exchange_configuration.available_required_time_frames,
                real_time_time_frames=exchange_configuration.real_time_time_frames,
                config_by_evaluator=self.config_by_tentacle,
            )

    async def _create_service_feeds(self):
        for feed in self.service_feeds:
            if not await service_api.start_service_feed(feed, False, {}):
                self.logger.error(f"Failed to start {feed.get_name()}. Evaluators requiring this service feed "
                                  f"might not work properly")

    async def _init_backtesting(self):
        if self.backtesting_data:
            self.backtesting_data.reset_cached_indexes()
        self.backtesting = await backtesting_api.initialize_backtesting(
            self.backtesting_config,
            exchange_ids=self.exchange_manager_ids,
            matrix_id=self.matrix_id,
            data_files=self.backtesting_files,
            importers_by_data_file=self.backtesting_data.importers_by_data_file if self.backtesting_data else None,
            backtest_data=self.backtesting_data,
            bot_id=self.bot_id,
        )
        if self.run_on_all_available_time_frames:
            self.backtesting_config[evaluator_constants.CONFIG_FORCED_TIME_FRAME] = [
                tf
                for tf in self.backtesting.importers[0].time_frames
            ]

    async def _configure_backtesting_time_window(self):
        # modify_backtesting_channels before creating exchanges as they require the current backtesting time to
        # initialize
        min_timestamp, max_timestamp = await backtesting_api.adapt_backtesting_channels(
            self.backtesting,
            self.backtesting_config,
            importers.ExchangeDataImporter,
            run_on_common_part_only=self.run_on_common_part_only,
            start_timestamp=self.start_timestamp,
            end_timestamp=self.end_timestamp
        )
        return min_timestamp, max_timestamp

    async def _fetch_backtesting_extra_data_if_any(
        self, min_timestamp: float, max_timestamp: float
    ):
        if not self.evaluators:
            return
        handled_classes = set()
        coros = []
        for evaluator in list_util.flatten_list(self.evaluators):
            if evaluator and evaluator.get_name() not in handled_classes:
                if evaluator.get_signals_history_type() == commons_enums.SignalHistoryTypes.GPT:
                    coros.append(self._fetch_gpt_history(evaluator, min_timestamp, max_timestamp))
                handled_classes.add(evaluator.get_name())
        if coros:
            self.has_fetched_data = True
            await asyncio.gather(*coros)

    async def _fetch_gpt_history(self, evaluator, min_timestamp: float, max_timestamp: float):
        # prevent circular import
        import tentacles.Services.Services_bases.gpt_service as gpt_service
        service = await service_api.get_service(gpt_service.GPTService, True, self.services_config)
        version = evaluator.get_version()
        for exchange_id in self.exchange_manager_ids:
            exchange_configuration = trading_api.get_exchange_configuration_from_exchange_id(exchange_id)
            exchange_name = trading_api.get_exchange_name(
                trading_api.get_exchange_manager_from_exchange_id(exchange_id)
            )
            await service.fetch_gpt_history(
                exchange_name,
                [str(symbol) for symbol in self.symbols_to_create_exchange_classes.get(exchange_name, [])],
                exchange_configuration.available_required_time_frames,
                version,
                min_timestamp,
                max_timestamp
            )

    async def clear_fetched_data(self):
        if self.has_fetched_data:
            # prevent circular import
            import tentacles.Services.Services_bases.gpt_service as gpt_service
            (await service_api.get_service(gpt_service.GPTService, True, self.services_config)).clear_signal_history()

    async def _init_exchanges(self):
        for exchange_class_string in self.symbols_to_create_exchange_classes.keys():
            is_future = self.exchange_type_by_exchange[exchange_class_string] == \
                        commons_constants.CONFIG_EXCHANGE_FUTURE
            exchange_builder = trading_api.create_exchange_builder(self.backtesting_config, exchange_class_string) \
                .has_matrix(self.matrix_id) \
                .use_tentacles_setup_config(self.tentacles_setup_config) \
                .use_trading_config_by_trading_mode(self.config_by_tentacle) \
                .has_forced_markets(self.backtesting_data.forced_markets if self.backtesting_data else None) \
                .set_bot_id(self.bot_id) \
                .is_simulated() \
                .is_rest_only() \
                .is_backtesting(self.backtesting) \
                .is_future(is_future, self.futures_contract_type) \
                .enable_storage(self.enable_storage)
            try:
                await exchange_builder.build()
            finally:
                # always save exchange manager ids and backtesting instances
                self.exchange_manager_ids.append(trading_api.get_exchange_manager_id(exchange_builder.exchange_manager))

    async def start_loggers(self):
        await self.start_exchange_loggers()
        await logger.init_evaluator_chan_logger(self.matrix_id)

    async def start_exchange_loggers(self):
        for exchange_manager_id in self.exchange_manager_ids:
            await logger.init_exchange_chan_logger(exchange_manager_id)


def _get_remaining_object_error(obj, expected, actual):
    error = f"too many remaining {obj.__name__} instances: expected: {expected} actual {actual[0]}"
    for i in range(len(actual[1])):
        debug_info = ""
        if isinstance(actual[1][i], exchanges.ExchangeManager):
            debug_info = f" ({actual[1][i].debug_info})"
        error += f"{sys.getrefcount(actual[1][i])} references on {actual[1][i]} {debug_info}"
        return error
